#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_utils.fixture import uuidsentinel as uuids

from placement import exception
from placement.objects import project as project_obj
from placement.tests.functional.db import test_base as tb


class ProjectTestCase(tb.PlacementDbBaseTestCase):
    def test_non_existing_project(self):
        self.assertRaises(
            exception.ProjectNotFound, project_obj.Project.get_by_external_id,
            self.ctx, uuids.non_existing_project)

    def test_create_and_get(self):
        p = project_obj.Project(self.ctx, external_id='another-project')
        p.create()
        p = project_obj.Project.get_by_external_id(self.ctx, 'another-project')
        # Project ID == 1 is fake-project created in setup
        self.assertEqual(2, p.id)
        self.assertRaises(exception.ProjectExists, p.create)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import alembic
from alembic import config as alembic_config
from alembic import migration as alembic_migration

from placement.db.sqlalchemy import models
from placement import db_api as placement_db


def get_engine():
    return placement_db.get_placement_engine()


def _alembic_config():
    path = os.path.join(os.path.dirname(__file__), "alembic.ini")
    config = alembic_config.Config(path)
    return config


def create_schema(engine=None):
    """Create schema from models, without a migration."""
    base = models.BASE

    if engine is None:
        engine = get_engine()
    base.metadata.create_all(engine)


def version(config=None, engine=None):
    """Current database version.

    :returns: Database version
    :rtype: string
    """
    if engine is None:
        engine = get_engine()
    with engine.connect() as conn:
        context = alembic_migration.MigrationContext.configure(conn)
        return context.get_current_revision()


def upgrade(revision, config=None):
    """Used for upgrading database.

    :param version: Desired database version
    :type version: string
    """
    revision = revision or "head"
    config = config or _alembic_config()
    alembic.command.upgrade(config, revision)


def stamp(version, config=None):
    """Used for stamp the database version.

    :param version: Database version to stamp
    :type version: string
    """
    config = config or _alembic_config()
    alembic.command.stamp(config, version)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Placement API schemas for usage information."""

import copy

from placement.schemas import common


# Represents the allowed query string parameters to GET /usages
GET_USAGES_SCHEMA_1_9 = {
    "type": "object",
    "properties": {
        "project_id": {
            "type": "string",
            "minLength": 1,
            "maxLength": 255,
        },
        "user_id": {
            "type": "string",
            "minLength": 1,
            "maxLength": 255,
        },
    },
    "required": [
        "project_id"
    ],
    "additionalProperties": False,
}


# An optional consumer type was added to the usage dicts in this
# version of GET /usages.

GET_USAGES_SCHEMA_V1_38 = copy.deepcopy(GET_USAGES_SCHEMA_1_9)
GET_USAGES_SCHEMA_V1_38['properties']['consumer_type'] = {
    "type": "string",
    "pattern": common.CONSUMER_TYPE_GET_PATTERN,
    "minLength": 1,
    "maxLength": 255,
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for code in the trait handler that gabbi cannot easily cover."""


from unittest import mock

import microversion_parse
import webob

from placement import context
from placement import exception
from placement.handlers import trait
from placement.tests.unit import base


class TestTraitHandler(base.ContextTestCase):

    @mock.patch('placement.objects.trait.Trait.create')
    @mock.patch('placement.objects.trait.Trait.get_by_name')
    @mock.patch('placement.context.RequestContext.can')
    @mock.patch('placement.util.wsgi_path_item', return_value='CUSTOM_FOOBAR')
    def test_trait_create_ordering(
            self, mock_path, mock_can, mock_get_by_name, mock_create):
        """Test that we call Trait.create when get_by_name has a TraitNotFound
        and that if create can't create, we assume 204.
        """
        # The trait doesn't initially exist.
        mock_get_by_name.side_effect = exception.TraitNotFound(
            name='CUSTOM_FOOBAR')
        # But we fake that it does after first not finding it.
        mock_create.side_effect = exception.TraitExists(
            name='CUSTOM_FOOBAR')
        fake_context = context.RequestContext(
            user_id='fake', project_id='fake')

        req = webob.Request.blank('/traits/CUSTOM_FOOBAR')
        req.environ['placement.context'] = fake_context

        parse_version = microversion_parse.parse_version_string
        microversion = parse_version('1.15')
        microversion.max_version = parse_version('9.99')
        microversion.min_version = parse_version('1.0')
        req.environ['placement.microversion'] = microversion

        response = req.get_response(trait.put_trait)

        # Trait was assumed to exist.
        self.assertEqual('204 No Content', response.status)

        # We get a last modified header, even though we don't know the exact
        # create_at time (it is None on the Trait object and we fall back to
        # now)
        self.assertIn('last-modified', response.headers)

        # Confirm we checked to see if the trait exists, but the
        # side_effect happens
        mock_get_by_name.assert_called_once_with(fake_context, 'CUSTOM_FOOBAR')

        # Confirm we attempt to create the trait.
        mock_create.assert_called_once_with()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Base class and convenience utilities for functional placement tests."""

import copy

import os_resource_classes as orc
from oslo_utils.fixture import uuidsentinel as uuids
from oslo_utils import uuidutils

from placement import exception
from placement.objects import allocation as alloc_obj
from placement.objects import consumer as consumer_obj
from placement.objects import inventory as inv_obj
from placement.objects import project as project_obj
from placement.objects import resource_class as rc_obj
from placement.objects import resource_provider as rp_obj
from placement.objects import trait as trait_obj
from placement.objects import user as user_obj
from placement.tests.functional import base


DISK_INVENTORY = dict(
    total=200,
    reserved=10,
    min_unit=2,
    max_unit=5,
    step_size=1,
    allocation_ratio=1.0,
    resource_class=orc.DISK_GB
)

DISK_ALLOCATION = dict(
    consumer_id=uuids.disk_consumer,
    used=2,
    resource_class=orc.DISK_GB
)


def create_provider(context, name, *aggs, **kwargs):
    parent = kwargs.get('parent')
    uuid = kwargs.get('uuid', getattr(uuids, name))
    rp = rp_obj.ResourceProvider(context, name=name, uuid=uuid)
    if parent:
        rp.parent_provider_uuid = parent
    rp.create()
    if aggs:
        rp.set_aggregates(aggs)
    return rp


def ensure_rc(context, name):
    try:
        rc_obj.ResourceClass.get_by_name(context, name)
    except exception.NotFound:
        rc_obj.ResourceClass(context, name=name).create()


def add_inventory(rp, rc, total, **kwargs):
    ensure_rc(rp._context, rc)
    kwargs.setdefault('max_unit', total)
    inv = inv_obj.Inventory(rp._context, resource_provider=rp,
                            resource_class=rc, total=total, **kwargs)
    rp.add_inventory(inv)
    return inv


def set_traits(rp, *traits):
    tlist = []
    for tname in traits:
        try:
            trait = trait_obj.Trait.get_by_name(rp._context, tname)
        except exception.TraitNotFound:
            trait = trait_obj.Trait(rp._context, name=tname)
            trait.create()
        tlist.append(trait)
    rp.set_traits(tlist)
    return tlist


def ensure_consumer(ctx, user, project, consumer_id=None):
    # NOTE(efried): If not specified, use a random consumer UUID - we don't
    # want to override any existing allocations from the test case.
    consumer_id = consumer_id or uuidutils.generate_uuid()
    try:
        consumer = consumer_obj.Consumer.get_by_uuid(ctx, consumer_id)
    except exception.NotFound:
        consumer = consumer_obj.Consumer(
            ctx, uuid=consumer_id, user=user, project=project)
        consumer.create()
    return consumer


def set_allocation(ctx, rp, consumer, rc_used_dict):
    alloc = [
        alloc_obj.Allocation(
            resource_provider=rp, resource_class=rc,
            consumer=consumer, used=used)
        for rc, used in rc_used_dict.items()
    ]
    alloc_obj.replace_all(ctx, alloc)
    return alloc


def create_user_and_project(ctx, prefix='fake'):
    user = user_obj.User(ctx, external_id='%s-user' % prefix)
    user.create()
    proj = project_obj.Project(ctx, external_id='%s-project' % prefix)
    proj.create()
    return user, proj


class PlacementDbBaseTestCase(base.TestCase):

    def setUp(self):
        super(PlacementDbBaseTestCase, self).setUp()
        # we use context in some places and ctx in other. We should only use
        # context, but let's paper over that for now.
        self.ctx = self.context
        self.user_obj, self.project_obj = create_user_and_project(self.ctx)
        # For debugging purposes, populated by _create_provider and used by
        # _validate_allocation_requests to make failure results more readable.
        self.rp_uuid_to_name = {}
        self.rp_id_to_name = {}

    def _assert_traits(self, expected_traits, traits_objs):
        expected_traits.sort()
        traits = []
        for obj in traits_objs:
            traits.append(obj.name)
        traits.sort()
        self.assertEqual(expected_traits, traits)

    def _assert_traits_in(self, expected_traits, traits_objs):
        traits = [trait.name for trait in traits_objs]
        for expected in expected_traits:
            self.assertIn(expected, traits)

    def _create_provider(self, name, *aggs, **kwargs):
        rp = create_provider(self.ctx, name, *aggs, **kwargs)
        self.rp_uuid_to_name[rp.uuid] = name
        self.rp_id_to_name[rp.id] = name
        return rp

    def get_provider_id_by_name(self, name):
        rp_ids = [k for k, v in self.rp_id_to_name.items() if v == name]
        if not len(rp_ids) == 1:
            raise Exception
        return rp_ids[0]

    def allocate_from_provider(self, rp, rc, used, consumer_id=None,
                               consumer=None):
        if consumer is None:
            consumer = ensure_consumer(
                self.ctx, self.user_obj, self.project_obj, consumer_id)
        alloc_list = set_allocation(self.ctx, rp, consumer, {rc: used})
        return alloc_list

    def _make_allocation(self, inv_dict, alloc_dict):
        alloc_dict = copy.copy(alloc_dict)
        rp = self._create_provider('allocation_resource_provider')
        disk_inv = inv_obj.Inventory(resource_provider=rp, **inv_dict)
        rp.set_inventory([disk_inv])
        consumer_id = alloc_dict.pop('consumer_id')
        consumer = ensure_consumer(
            self.ctx, self.user_obj, self.project_obj, consumer_id)
        alloc = alloc_obj.Allocation(
            resource_provider=rp, consumer=consumer, **alloc_dict)
        alloc_obj.replace_all(self.ctx, [alloc])
        return rp, alloc

    def create_aggregate(self, agg_uuid):
        conn = self.placement_db.get_engine().connect()
        ins_stmt = rp_obj._AGG_TBL.insert().values(uuid=agg_uuid)
        with conn.begin():
            res = conn.execute(ins_stmt)
        agg_id = res.inserted_primary_key[0]
        return agg_id



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_policy import policy

from placement.policies import base


PREFIX = 'placement:resource_providers:aggregates:%s'
LIST = PREFIX % 'list'
UPDATE = PREFIX % 'update'
BASE_PATH = '/resource_providers/{uuid}/aggregates'

rules = [
    policy.DocumentedRuleDefault(
        LIST,
        base.ADMIN_OR_SERVICE,
        "List resource provider aggregates.",
        [
            {
                'method': 'GET',
                'path': BASE_PATH
            }
        ],
        scope_types=['project'],
    ),
    policy.DocumentedRuleDefault(
        UPDATE,
        base.ADMIN_OR_SERVICE,
        "Update resource provider aggregates.",
        [
            {
                'method': 'PUT',
                'path': BASE_PATH
            }
        ],
        scope_types=['project'],
    ),
]


def list_rules():
    return rules



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import os_resource_classes as orc
from oslo_utils.fixture import uuidsentinel
from oslo_utils import uuidutils

from placement.objects import consumer as c_obj
from placement.objects import consumer_type as ct_obj
from placement.objects import inventory as inv_obj
from placement.objects import usage as usage_obj
from placement.tests.functional.db import test_base as tb


class UsageListTestCase(tb.PlacementDbBaseTestCase):

    def test_get_all_null(self):
        for uuid in [uuidsentinel.rp_uuid_1, uuidsentinel.rp_uuid_2]:
            self._create_provider(uuid, uuid=uuid)

        usages = usage_obj.get_all_by_resource_provider_uuid(
            self.ctx, uuidsentinel.rp_uuid_1)
        self.assertEqual(0, len(usages))

    def test_get_all_one_allocation(self):
        db_rp, _ = self._make_allocation(tb.DISK_INVENTORY,
                                         tb.DISK_ALLOCATION)
        inv = inv_obj.Inventory(resource_provider=db_rp,
                                resource_class=orc.DISK_GB,
                                total=1024)
        db_rp.set_inventory([inv])

        usages = usage_obj.get_all_by_resource_provider_uuid(
            self.ctx, db_rp.uuid)
        self.assertEqual(1, len(usages))
        self.assertEqual(2, usages[0].usage)
        self.assertEqual(orc.DISK_GB,
                         usages[0].resource_class)

    def test_get_inventory_no_allocation(self):
        db_rp = self._create_provider('rp_no_inv')
        tb.add_inventory(db_rp, orc.DISK_GB, 1024)

        usages = usage_obj.get_all_by_resource_provider_uuid(
            self.ctx, db_rp.uuid)
        self.assertEqual(1, len(usages))
        self.assertEqual(0, usages[0].usage)
        self.assertEqual(orc.DISK_GB,
                         usages[0].resource_class)

    def test_get_all_multiple_inv(self):
        db_rp = self._create_provider('rp_no_inv')
        tb.add_inventory(db_rp, orc.DISK_GB, 1024)
        tb.add_inventory(db_rp, orc.VCPU, 24)

        usages = usage_obj.get_all_by_resource_provider_uuid(
            self.ctx, db_rp.uuid)
        self.assertEqual(2, len(usages))

    def test_get_by_unspecified_consumer_type(self):
        # This will add a consumer with a NULL consumer type and the default
        # project and user external_ids
        self._make_allocation(tb.DISK_INVENTORY, tb.DISK_ALLOCATION)

        # Verify we filter the project external_id correctly. Note: this will
        # also work if filtering is broken (if it's not filtering at all)
        usages = usage_obj.get_by_consumer_type(
            self.ctx, self.project_obj.external_id)
        self.assertEqual(1, len(usages))
        usage = usages[0]
        self.assertEqual('unknown', usage.consumer_type)
        self.assertEqual(1, usage.consumer_count)
        self.assertEqual(orc.DISK_GB, usage.resource_class)
        self.assertEqual(2, usage.usage)
        # Verify we get nothing back if we filter on a different project
        # external_id that does not exist (will not work if filtering is
        # broken)
        usages = usage_obj.get_by_consumer_type(self.ctx, 'BOGUS')
        self.assertEqual(0, len(usages))

    def test_get_by_specified_consumer_type(self):
        ct = ct_obj.ConsumerType(self.ctx, name='INSTANCE')
        ct.create()
        consumer_id = uuidutils.generate_uuid()
        c = c_obj.Consumer(self.ctx, uuid=consumer_id,
                           project=self.project_obj, user=self.user_obj,
                           consumer_type_id=ct.id)
        c.create()
        # This will add a consumer with the consumer type INSTANCE
        # and the default project and user external_ids
        da = copy.deepcopy(tb.DISK_ALLOCATION)
        da['consumer_id'] = c.uuid
        self._make_allocation(tb.DISK_INVENTORY, da)

        # Verify we filter the INSTANCE type correctly. Note: this will also
        # work if filtering is broken (if it's not filtering at all)
        usages = usage_obj.get_by_consumer_type(
            self.ctx, self.project_obj.external_id,
            consumer_type=ct.name)
        self.assertEqual(1, len(usages))
        usage = usages[0]
        self.assertEqual(ct.name, usage.consumer_type)
        self.assertEqual(1, usage.consumer_count)
        self.assertEqual(orc.DISK_GB, usage.resource_class)
        self.assertEqual(2, usage.usage)
        # Verify we get nothing back if we filter on a different consumer
        # type that does not exist (will not work if filtering is broken)
        usages = usage_obj.get_by_consumer_type(
            self.ctx, self.project_obj.external_id,
            consumer_type='BOGUS')
        self.assertEqual(0, len(usages))

    def test_get_by_specified_consumer_type_with_user(self):
        ct = ct_obj.ConsumerType(self.ctx, name='INSTANCE')
        ct.create()
        consumer_id = uuidutils.generate_uuid()
        c = c_obj.Consumer(self.ctx, uuid=consumer_id,
                           project=self.project_obj, user=self.user_obj,
                           consumer_type_id=ct.id)
        c.create()
        # This will add a consumer with the consumer type INSTANCE
        # and the default project and user external_ids
        da = copy.deepcopy(tb.DISK_ALLOCATION)
        da['consumer_id'] = c.uuid
        db_rp, _ = self._make_allocation(tb.DISK_INVENTORY, da)

        # Verify we filter the user external_id correctly. Note: this will also
        # work if filtering is broken (if it's not filtering at all)
        usages = usage_obj.get_by_consumer_type(
            self.ctx, self.project_obj.external_id,
            user_id=self.user_obj.external_id,
            consumer_type=ct.name)
        self.assertEqual(1, len(usages))
        usage = usages[0]
        self.assertEqual(ct.name, usage.consumer_type)
        self.assertEqual(1, usage.consumer_count)
        self.assertEqual(orc.DISK_GB, usage.resource_class)
        self.assertEqual(2, usage.usage)
        # Verify we get nothing back if we filter on a different user
        # external_id that does not exist (will not work if filtering is
        # broken)
        usages = usage_obj.get_by_consumer_type(
            self.ctx, self.project_obj.external_id,
            user_id='BOGUS',
            consumer_type=ct.name)
        self.assertEqual(0, len(usages))

    def test_get_by_all_consumer_type(self):
        # This will add a consumer with the default consumer type UNKNOWN
        db_rp, _ = self._make_allocation(tb.DISK_INVENTORY,
                                         tb.DISK_ALLOCATION)
        # Make another allocation with a different consumer type
        ct = ct_obj.ConsumerType(self.ctx, name='FOO')
        ct.create()
        consumer_id = uuidutils.generate_uuid()
        c = c_obj.Consumer(self.ctx, uuid=consumer_id,
                           project=self.project_obj, user=self.user_obj,
                           consumer_type_id=ct.id)
        c.create()
        self.allocate_from_provider(db_rp, orc.DISK_GB, 2, consumer=c)

        # Verify we get usages back for both consumer types with 'all'
        usages = usage_obj.get_by_consumer_type(
            self.ctx, self.project_obj.external_id, consumer_type='all')
        self.assertEqual(1, len(usages))
        usage = usages[0]
        self.assertEqual('all', usage.consumer_type)
        self.assertEqual(2, usage.consumer_count)
        self.assertEqual(orc.DISK_GB, usage.resource_class)
        self.assertEqual(4, usage.usage)

    def test_get_by_unused_consumer_type(self):
        # This will add a consumer with the default consumer type UNKNOWN
        self._make_allocation(tb.DISK_INVENTORY, tb.DISK_ALLOCATION)

        usages = usage_obj.get_by_consumer_type(
            self.ctx, self.project_obj.external_id, consumer_type='EMPTY')
        self.assertEqual(0, len(usages))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Placement API handlers for getting allocation candidates."""

import collections

from oslo_serialization import jsonutils
from oslo_utils import encodeutils
from oslo_utils import timeutils
import webob

from placement import exception
from placement import lib
from placement import microversion
from placement.objects import allocation_candidate as ac_obj
from placement.policies import allocation_candidate as policies
from placement.schemas import allocation_candidate as schema
from placement import util
from placement import wsgi_wrapper

# The microversions at which the schema used to validate
# query parameters to GET /allocation_candidates differs.
_GET_SCHEMA_MICROVERSIONS = [
    (1, 36), (1, 35), (1, 33), (1, 31), (1, 25), (1, 21), (1, 17), (1, 16)
]


def _transform_allocation_requests_dict(alloc_reqs, want_version):
    """Turn supplied list of AllocationRequest objects into a list of
    allocations dicts keyed by resource provider uuid of resources involved
    in the allocation request. The returned results are intended to be used
    as the body of a PUT /allocations/{consumer_uuid} HTTP request at
    micoversion 1.12 (and beyond). The JSON objects look like the following:

    [
        {
            "allocations": {
                $rp_uuid1: {
                    "resources": {
                        "MEMORY_MB": 512
                        ...
                    }
                },
                $rp_uuid2: {
                    "resources": {
                        "DISK_GB": 1024
                        ...
                    }
                }
            },
            # If microversion >=1.34 then map suffixes to providers.
            "mappings": {
                "_COMPUTE": [$rp_uuid2],
                "": [$rp_uuid1]

            },
        },
        ...
    ]
    """
    results = []

    for ar in alloc_reqs:
        # A default dict of {$rp_uuid: "resources": {})
        rp_resources = collections.defaultdict(lambda: dict(resources={}))
        for rr in ar.resource_requests:
            res_dict = rp_resources[rr.resource_provider.uuid]['resources']
            res_dict[rr.resource_class] = rr.amount
        result = dict(allocations=rp_resources)
        if want_version.matches((1, 34)):
            result['mappings'] = ar.mappings
        results.append(result)

    return results


def _transform_allocation_requests_list(alloc_reqs):
    """Turn supplied list of AllocationRequest objects into a list of dicts of
    resources involved in the allocation request. The returned results is
    intended to be able to be used as the body of a PUT
    /allocations/{consumer_uuid} HTTP request, prior to microversion 1.12,
    so therefore we return a list of JSON objects that looks like the
    following:

    [
        {
            "allocations": [
                {
                    "resource_provider": {
                        "uuid": $rp_uuid,
                    }
                    "resources": {
                        $resource_class: $requested_amount, ...
                    },
                }, ...
            ],
        }, ...
    ]
    """
    results = []
    for ar in alloc_reqs:
        provider_resources = collections.defaultdict(dict)
        for rr in ar.resource_requests:
            res_dict = provider_resources[rr.resource_provider.uuid]
            res_dict[rr.resource_class] = rr.amount

        allocs = [
            {
                "resource_provider": {
                    "uuid": rp_uuid,
                },
                "resources": resources,
            } for rp_uuid, resources in provider_resources.items()
        ]
        alloc = {
            "allocations": allocs
        }
        results.append(alloc)
    return results


def _transform_provider_summaries(p_sums, requests, want_version):
    """Turn supplied list of ProviderSummary objects into a dict, keyed by
    resource provider UUID, of dicts of provider and inventory information.
    The traits only show up when `want_version` is 1.17 or newer. All the
    resource classes are shown when `want_version` is 1.27 or newer while
    only requested resources are included in the `provider_summaries`
    for older versions. The parent and root provider uuids only show up
    when `want_version` is 1.29 or newer.

    {
       RP_UUID_1: {
           'resources': {
              'DISK_GB': {
                'capacity': 100,
                'used': 0,
              },
              'VCPU': {
                'capacity': 4,
                'used': 0,
              }
           },
           # traits shows up from microversion 1.17
           'traits': [
                'HW_CPU_X86_AVX512F',
                'HW_CPU_X86_AVX512CD'
           ]
           # parent/root provider uuids show up from microversion 1.29
           parent_provider_uuid: null,
           root_provider_uuid: RP_UUID_1
       },
       RP_UUID_2: {
           'resources': {
              'DISK_GB': {
                'capacity': 100,
                'used': 0,
              },
              'VCPU': {
                'capacity': 4,
                'used': 0,
              }
           },
           # traits shows up from microversion 1.17
           'traits': [
                'HW_NIC_OFFLOAD_TSO',
                'HW_NIC_OFFLOAD_GRO'
           ],
           # parent/root provider uuids show up from microversion 1.29
           parent_provider_uuid: null,
           root_provider_uuid: RP_UUID_2
       }
    }
    """
    include_traits = want_version.matches((1, 17))
    include_all_resources = want_version.matches((1, 27))
    enable_nested_providers = want_version.matches((1, 29))

    ret = {}
    requested_resources = set()

    for requested_group in requests.values():
        requested_resources |= set(requested_group.resources)

    # if include_all_resources is false, only requested resources are
    # included in the provider_summaries.
    for ps in p_sums:
        resources = {
            psr.resource_class: {
                'capacity': psr.capacity,
                'used': psr.used,
            } for psr in ps.resources if (
                include_all_resources or
                psr.resource_class in requested_resources)
        }

        ret[ps.resource_provider.uuid] = {'resources': resources}

        if include_traits:
            ret[ps.resource_provider.uuid]['traits'] = ps.traits

        if enable_nested_providers:
            ret[ps.resource_provider.uuid]['parent_provider_uuid'] = (
                ps.resource_provider.parent_provider_uuid)
            ret[ps.resource_provider.uuid]['root_provider_uuid'] = (
                ps.resource_provider.root_provider_uuid)

    return ret


def _transform_allocation_candidates(alloc_cands, requests, want_version):
    """Turn supplied AllocationCandidates object into a dict containing
    allocation requests and provider summaries.

    {
        'allocation_requests': <ALLOC_REQUESTS>,
        'provider_summaries': <PROVIDER_SUMMARIES>,
    }
    """
    if want_version.matches((1, 12)):
        a_reqs = _transform_allocation_requests_dict(
            alloc_cands.allocation_requests, want_version)
    else:
        a_reqs = _transform_allocation_requests_list(
            alloc_cands.allocation_requests)

    p_sums = _transform_provider_summaries(
        alloc_cands.provider_summaries, requests, want_version)

    return {
        'allocation_requests': a_reqs,
        'provider_summaries': p_sums,
    }


def _get_schema(want_version):
    """Calculate the desired query parameter schema for
    list_allocation_candidates.
    """
    for maj, min in _GET_SCHEMA_MICROVERSIONS:
        if want_version.matches((maj, min)):
            return getattr(schema, 'GET_SCHEMA_%d_%d' % (maj, min))

    return schema.GET_SCHEMA_1_10


@wsgi_wrapper.PlacementWsgify
@microversion.version_handler('1.10')
@util.check_accept('application/json')
def list_allocation_candidates(req):
    """GET a JSON object with a list of allocation requests and a JSON object
    of provider summary objects

    On success return a 200 and an application/json body representing
    a collection of allocation requests and provider summaries
    """
    context = req.environ['placement.context']
    context.can(policies.LIST)
    want_version = req.environ[microversion.MICROVERSION_ENVIRON]
    get_schema = _get_schema(want_version)
    util.validate_query_params(req, get_schema)

    rqparams = lib.RequestWideParams.from_request(req)
    groups = lib.RequestGroup.dict_from_request(req, rqparams)

    if not rqparams.group_policy:
        # group_policy is required if more than one numbered request group was
        # specified.
        if len([rg for rg in groups.values() if rg.use_same_provider]) > 1:
            raise webob.exc.HTTPBadRequest(
                'The "group_policy" parameter is required when specifying '
                'more than one "resources{N}" parameter.')

    # We can't be aware of nested architecture with old microversions
    nested_aware = want_version.matches((1, 29))

    try:
        cands = ac_obj.AllocationCandidates.get_by_requests(
            context, groups, rqparams, nested_aware=nested_aware)
    except exception.ResourceClassNotFound as exc:
        raise webob.exc.HTTPBadRequest(
            'Invalid resource class in resources parameter: %(error)s' %
            {'error': exc})
    except exception.TraitNotFound as exc:
        raise webob.exc.HTTPBadRequest(str(exc))

    response = req.response
    trx_cands = _transform_allocation_candidates(cands, groups, want_version)
    json_data = jsonutils.dumps(trx_cands)
    response.body = encodeutils.to_utf8(json_data)
    response.content_type = 'application/json'
    if want_version.matches((1, 15)):
        response.cache_control = 'no-cache'
        response.last_modified = timeutils.utcnow(with_timezone=True)
    return response



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import os_resource_classes as orc
from oslo_utils.fixture import uuidsentinel as uuids
from oslo_utils import timeutils

from placement.objects import inventory
from placement.objects import resource_provider
from placement.tests.unit.objects import base


_RESOURCE_CLASS_NAME = 'DISK_GB'
_RESOURCE_CLASS_ID = 2
_RESOURCE_PROVIDER_ID = 1
_RESOURCE_PROVIDER_UUID = uuids.resource_provider

VCPU_ID = orc.STANDARDS.index(
    orc.VCPU)

_INVENTORY_ID = 2
_INVENTORY_DB = {
    'id': _INVENTORY_ID,
    'resource_provider_id': _RESOURCE_PROVIDER_ID,
    'resource_class_id': _RESOURCE_CLASS_ID,
    'total': 16,
    'reserved': 2,
    'min_unit': 1,
    'max_unit': 8,
    'step_size': 1,
    'allocation_ratio': 1.0,
    'updated_at': None,
    'created_at': timeutils.utcnow(with_timezone=True),
}


class TestInventoryNoDB(base.TestCase):

    @mock.patch('placement.objects.inventory._get_inventory_by_provider_id')
    def test_get_all_by_resource_provider(self, mock_get):
        expected = [dict(_INVENTORY_DB,
                         resource_provider_id=_RESOURCE_PROVIDER_ID),
                    dict(_INVENTORY_DB,
                         id=_INVENTORY_DB['id'] + 1,
                         resource_provider_id=_RESOURCE_PROVIDER_ID)]
        mock_get.return_value = expected
        rp = resource_provider.ResourceProvider(self.context,
                                                id=_RESOURCE_PROVIDER_ID,
                                                uuid=_RESOURCE_PROVIDER_UUID)
        objs = inventory.get_all_by_resource_provider(self.context, rp)
        self.assertEqual(2, len(objs))
        self.assertEqual(_INVENTORY_DB['id'], objs[0].id)
        self.assertEqual(_INVENTORY_DB['id'] + 1, objs[1].id)
        self.assertEqual(_RESOURCE_PROVIDER_ID, objs[0].resource_provider.id)

    def test_set_defaults(self):
        rp = resource_provider.ResourceProvider(self.context,
                                                id=_RESOURCE_PROVIDER_ID,
                                                uuid=_RESOURCE_PROVIDER_UUID)
        kwargs = dict(resource_provider=rp,
                      resource_class=_RESOURCE_CLASS_NAME,
                      total=16)
        inv = inventory.Inventory(self.context, **kwargs)

        self.assertEqual(0, inv.reserved)
        self.assertEqual(1, inv.min_unit)
        self.assertEqual(1, inv.max_unit)
        self.assertEqual(1, inv.step_size)
        self.assertEqual(1.0, inv.allocation_ratio)

    def test_capacity(self):
        rp = resource_provider.ResourceProvider(self.context,
                                                id=_RESOURCE_PROVIDER_ID,
                                                uuid=_RESOURCE_PROVIDER_UUID)
        kwargs = dict(resource_provider=rp,
                      resource_class=_RESOURCE_CLASS_NAME,
                      total=16,
                      reserved=16)
        inv = inventory.Inventory(self.context, **kwargs)

        self.assertEqual(0, inv.capacity)
        inv.reserved = 15
        self.assertEqual(1, inv.capacity)
        inv.allocation_ratio = 2.0
        self.assertEqual(2, inv.capacity)


class TestListOfInventory(base.TestCase):

    def test_find(self):
        rp = resource_provider.ResourceProvider(
            self.context, uuid=uuids.rp_uuid)
        inv_list = [
            inventory.Inventory(
                resource_provider=rp,
                resource_class=orc.VCPU,
                total=24),
            inventory.Inventory(
                resource_provider=rp,
                resource_class=orc.MEMORY_MB,
                total=10240),
        ]

        found = inventory.find(inv_list, orc.MEMORY_MB)
        self.assertIsNotNone(found)
        self.assertEqual(10240, found.total)

        found = inventory.find(inv_list, orc.VCPU)
        self.assertIsNotNone(found)
        self.assertEqual(24, found.total)

        found = inventory.find(inv_list, orc.DISK_GB)
        self.assertIsNone(found)

        # Try an integer resource class identifier...
        self.assertRaises(ValueError, inventory.find, inv_list, VCPU_ID)

        # Use an invalid string...
        self.assertIsNone(inventory.find(inv_list, 'HOUSE'))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Unit tests for code in the resource provider handler that gabbi isn't covering.
"""

from unittest import mock

import microversion_parse
from oslo_db import exception as db_exc
import webob

from placement import context
from placement.handlers import resource_provider
from placement.tests.unit import base


class TestAggregateHandlerErrors(base.ContextTestCase):

    @mock.patch('placement.context.RequestContext.can', new=mock.Mock())
    def _test_duplicate_error_parsing_mysql(self, key):
        fake_context = context.RequestContext(
            user_id='fake', project_id='fake')

        req = webob.Request.blank(
            '/resource_providers',
            method='POST',
            content_type='application/json')
        req.body = b'{"name": "foobar"}'
        req.environ['placement.context'] = fake_context

        parse_version = microversion_parse.parse_version_string
        microversion = parse_version('1.15')
        microversion.max_version = parse_version('9.99')
        microversion.min_version = parse_version('1.0')
        req.environ['placement.microversion'] = microversion

        with mock.patch(
            'placement.objects.resource_provider.ResourceProvider.create',
            side_effect=db_exc.DBDuplicateEntry(columns=[key]),
        ):
            response = req.get_response(
                resource_provider.create_resource_provider)

        self.assertEqual('409 Conflict', response.status)
        self.assertIn(
            'Conflicting resource provider name: foobar already exists.',
            response.text)

    def test_duplicate_error_parsing_mysql_5x(self):
        """Ensure we parse the correct column on MySQL 5.x.

        On MySQL 5.x, DBDuplicateEntry.columns will contain the name of the
        column causing the integrity error.
        """
        self._test_duplicate_error_parsing_mysql('name')

    def test_duplicate_error_parsing_mysql_8x(self):
        """Ensure we parse the correct column on MySQL 5.x.

        On MySQL 5.x, DBDuplicateEntry.columns will contain the name of the
        constraint causing the integrity error.
        """
        self._test_duplicate_error_parsing_mysql(
            'uniq_resource_providers0name')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Drop redundant indexes for unique constraints

Revision ID: a082b8bb98d0
Revises: 422ece571366
Create Date: 2022-09-09 15:52:21.644040

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'a082b8bb98d0'
down_revision = '422ece571366'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index('inventories_resource_provider_id_idx',
                  table_name='inventories')
    op.drop_index('inventories_resource_provider_resource_class_idx',
                  table_name='inventories')
    op.drop_index('ix_placement_aggregates_uuid',
                  table_name='placement_aggregates')
    op.drop_index('resource_providers_name_idx',
                  table_name='resource_providers')
    op.drop_index('resource_providers_uuid_idx',
                  table_name='resource_providers')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Placement API schemas for resource classes."""

import copy

from placement.schemas import common


POST_RC_SCHEMA_V1_2 = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "pattern": common.CUSTOM_RC_PATTERN,
            "maxLength": 255,
        },
    },
    "required": [
        "name"
    ],
    "additionalProperties": False,
}
PUT_RC_SCHEMA_V1_2 = copy.deepcopy(POST_RC_SCHEMA_V1_2)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from oslo_config import cfg
from oslo_config import fixture as config_fixture

from placement import conf


class TestPlacementDBConf(testtools.TestCase):
    """Test cases for Placement DB Setup."""

    def setUp(self):
        super(TestPlacementDBConf, self).setUp()
        config = cfg.ConfigOpts()
        self.conf_fixture = self.useFixture(config_fixture.Config(config))
        conf.register_opts(config)

    def test_missing_config_raises(self):
        """Not setting [placement_database]/connection is an error."""
        exc = self.assertRaises(
            cfg.RequiredOptError, self.conf_fixture.conf,
            [], default_config_files=[])
        self.assertIn(
            'option connection in group [placement_database]',
            str(exc))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Utility methods for getting allocation candidates."""

import collections


RPCandidate = collections.namedtuple('RPCandidates', 'id root_id rc_id')


class RPCandidateList(object):
    """Helper class to manage allocation candidate resource providers list,
    RPCandidates, which consists of three-tuples with the first element being
    the resource provider ID, the second element being the root provider ID
    and the third being resource class ID.
    """

    def __init__(self, rp_candidates=None):
        self.rp_candidates = rp_candidates or set()

    def __len__(self):
        return len(self.rp_candidates)

    def __bool__(self):
        return bool(len(self))

    def __nonzero__(self):
        return self.__bool__()

    def merge_common_trees(self, other):
        """Merge two RPCandidateLists by OR'ing the two list of candidates
        and if the tree is not in both RPCandidateLists, we exclude resource
        providers in that tree. This is used to get trees that can satisfy
        all requested resource.
        """
        if not self:
            self.rp_candidates = other.rp_candidates
        elif not other:
            pass
        else:
            trees_in_both = self.trees & other.trees
            self.rp_candidates |= other.rp_candidates
            self.filter_by_tree(trees_in_both)

    def add_rps(self, rps, rc_id):
        """Add given resource providers to the candidate list.

        :param rps: tuples of (resource provider ID, anchor root provider ID)
        :param rc_id: ID of the class of resource provided by these resource
                      providers
        """
        self.rp_candidates |= set(
            RPCandidate(id=rp[0], root_id=rp[1], rc_id=rc_id) for rp in rps)

    def filter_by_tree(self, tree_root_ids):
        """Filter the candidates by given trees"""
        self.rp_candidates = set(
            p for p in self.rp_candidates if p.root_id in tree_root_ids)

    def filter_by_rp(self, rptuples):
        """Filter the candidates by given resource provider"""
        self.rp_candidates = set(
            p for p in self.rp_candidates if (p.id, p.root_id) in rptuples)

    def filter_by_rp_or_tree(self, rp_ids):
        """Filter the candidates out if neither itself nor its root is in
        given resource providers
        """
        self.rp_candidates = set(
            p for p in self.rp_candidates if set([p.id, p.root_id]) & rp_ids)

    def filter_by_rp_nor_tree(self, rp_ids):
        """Filter the candidates out if either itself or its root is in
        given resource providers
        """
        self.rp_candidates = set(
            p for p in self.rp_candidates if not (
                set([p.id, p.root_id]) & rp_ids))

    @property
    def rps(self):
        """Returns a set of IDs of nominated resource providers"""
        return set(p.id for p in self.rp_candidates)

    @property
    def trees(self):
        """Returns a set of nominated trees each of which are expressed by
        the root provider ID
        """
        return set(p.root_id for p in self.rp_candidates)

    @property
    def all_rps(self):
        """Returns a set of IDs of all involved resource providers"""
        return (self.rps | self.trees)

    @property
    def rps_info(self):
        return self.rp_candidates



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db import exception as db_exc
import sqlalchemy as sa

from placement.db.sqlalchemy import models
from placement import db_api
from placement import exception

PROJECT_TBL = models.Project.__table__


@db_api.placement_context_manager.writer
def ensure_incomplete_project(ctx):
    """Ensures that a project record is created for the "incomplete consumer
    project". Returns the internal ID of that record.
    """
    incomplete_id = ctx.config.placement.incomplete_consumer_project_id
    sel = sa.select(PROJECT_TBL.c.id).where(
        PROJECT_TBL.c.external_id == incomplete_id)
    res = ctx.session.execute(sel).fetchone()
    if res:
        return res[0]
    ins = PROJECT_TBL.insert().values(external_id=incomplete_id)
    res = ctx.session.execute(ins)
    return res.inserted_primary_key[0]


@db_api.placement_context_manager.reader
def _get_project_by_external_id(ctx, external_id):
    projects = sa.alias(PROJECT_TBL, name="p")
    sel = sa.select(
        projects.c.id,
        projects.c.external_id,
        projects.c.updated_at,
        projects.c.created_at,
    )
    sel = sel.where(projects.c.external_id == external_id)
    res = ctx.session.execute(sel).fetchone()
    if not res:
        raise exception.ProjectNotFound(external_id=external_id)

    return dict(res._mapping)


class Project(object):

    def __init__(self, context, id=None, external_id=None, updated_at=None,
                 created_at=None):
        self._context = context
        self.id = id
        self.external_id = external_id
        self.updated_at = updated_at
        self.created_at = created_at

    @staticmethod
    def _from_db_object(ctx, target, source):
        target._context = ctx
        target.id = source['id']
        target.external_id = source['external_id']
        target.updated_at = source['updated_at']
        target.created_at = source['created_at']
        return target

    @classmethod
    def get_by_external_id(cls, ctx, external_id):
        res = _get_project_by_external_id(ctx, external_id)
        return cls._from_db_object(ctx, cls(ctx), res)

    def create(self):
        @db_api.placement_context_manager.writer
        def _create_in_db(ctx):
            db_obj = models.Project(external_id=self.external_id)
            try:
                db_obj.save(ctx.session)
            except db_exc.DBDuplicateEntry:
                raise exception.ProjectExists(external_id=self.external_id)
            self._from_db_object(ctx, self, db_obj)
        _create_in_db(self._context)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the functions used by the placement API handlers."""

from unittest import mock

import microversion_parse
from oslo_utils.fixture import uuidsentinel
import routes
import testtools
import webob

from placement import handler
from placement.handlers import root
from placement import microversion


# Used in tests below
def start_response(*args, **kwargs):
    pass


def _environ(path='/moo', method='GET'):
    return {
        'PATH_INFO': path,
        'REQUEST_METHOD': method,
        'SERVER_NAME': 'example.com',
        'SERVER_PORT': '80',
        'wsgi.url_scheme': 'http',
        # The microversion version value is not used, but it
        # needs to be set to avoid a KeyError.
        microversion.MICROVERSION_ENVIRON: microversion_parse.Version(1, 12),
    }


class DispatchTest(testtools.TestCase):

    def setUp(self):
        super(DispatchTest, self).setUp()
        self.mapper = routes.Mapper()
        self.route_handler = mock.MagicMock()

    def test_no_match_null_map(self):
        self.assertRaises(webob.exc.HTTPNotFound,
                          handler.dispatch,
                          _environ(), start_response,
                          self.mapper)

    def test_no_match_with_map(self):
        self.mapper.connect('/foobar', action='hello')
        self.assertRaises(webob.exc.HTTPNotFound,
                          handler.dispatch,
                          _environ(), start_response,
                          self.mapper)

    def test_simple_match(self):
        self.mapper.connect('/foobar', action=self.route_handler,
                            conditions=dict(method=['GET']))
        environ = _environ(path='/foobar')
        handler.dispatch(environ, start_response, self.mapper)
        self.route_handler.assert_called_with(environ, start_response)

    def test_simple_match_routing_args(self):
        self.mapper.connect('/foobar/{id}', action=self.route_handler,
                            conditions=dict(method=['GET']))
        environ = _environ(path='/foobar/%s' % uuidsentinel.foobar)
        handler.dispatch(environ, start_response, self.mapper)
        self.route_handler.assert_called_with(environ, start_response)
        self.assertEqual(uuidsentinel.foobar,
                         environ['wsgiorg.routing_args'][1]['id'])


class MapperTest(testtools.TestCase):

    def setUp(self):
        super(MapperTest, self).setUp()
        declarations = {
            '/hello': {'GET': 'hello'}
        }
        self.mapper = handler.make_map(declarations)

    def test_no_match(self):
        environ = _environ(path='/cow')
        self.assertIsNone(self.mapper.match(environ=environ))

    def test_match(self):
        environ = _environ(path='/hello')
        action = self.mapper.match(environ=environ)['action']
        self.assertEqual('hello', action)

    def test_405_methods(self):
        environ = _environ(path='/hello', method='POST')
        result = self.mapper.match(environ=environ)
        self.assertEqual(handler.handle_405, result['action'])
        self.assertEqual('GET', result['_methods'])

    def test_405_headers(self):
        environ = _environ(path='/hello', method='POST')
        global headers, status
        headers = status = None

        def local_start_response(*args, **kwargs):
            global headers, status
            status = args[0]
            headers = {header[0]: header[1] for header in args[1]}

        handler.dispatch(environ, local_start_response, self.mapper)
        allow_header = headers['allow']
        self.assertEqual('405 Method Not Allowed', status)
        self.assertEqual('GET', allow_header)
        # PEP 3333 requires that headers be whatever the native str
        # is in that version of Python. Never unicode.
        self.assertEqual(str, type(allow_header))


class PlacementLoggingTest(testtools.TestCase):

    @mock.patch("placement.handler.LOG")
    def test_404_no_error_log(self, mocked_log):
        environ = _environ(path='/hello', method='GET')
        config = mock.MagicMock()
        context_mock = mock.Mock()
        context_mock.to_policy_values.return_value = {'roles': ['admin']}
        environ['placement.context'] = context_mock
        app = handler.PlacementHandler(config=config)
        self.assertRaises(webob.exc.HTTPNotFound,
                          app, environ, start_response)
        mocked_log.error.assert_not_called()
        mocked_log.exception.assert_not_called()


class DeclarationsTest(testtools.TestCase):

    def setUp(self):
        super(DeclarationsTest, self).setUp()
        self.mapper = handler.make_map(handler.ROUTE_DECLARATIONS)

    def test_root_slash_match(self):
        environ = _environ(path='/')
        result = self.mapper.match(environ=environ)
        self.assertEqual(root.home, result['action'])

    def test_root_empty_match(self):
        environ = _environ(path='')
        result = self.mapper.match(environ=environ)
        self.assertEqual(root.home, result['action'])


class ContentHeadersTest(testtools.TestCase):

    def setUp(self):
        super(ContentHeadersTest, self).setUp()
        self.environ = _environ(path='/')
        config = mock.MagicMock()
        self.environ['placement.context'] = mock.MagicMock()
        self.app = handler.PlacementHandler(config=config)

    def test_no_content_type(self):
        self.environ['CONTENT_LENGTH'] = '10'
        self.assertRaisesRegex(webob.exc.HTTPBadRequest,
                               "content-type header required when "
                               "content-length > 0", self.app,
                               self.environ, start_response)

    def test_non_integer_content_length(self):
        self.environ['CONTENT_LENGTH'] = 'foo'
        self.assertRaisesRegex(webob.exc.HTTPBadRequest,
                               "content-length header must be an integer",
                               self.app, self.environ, start_response)

    def test_empty_content_type(self):
        self.environ['CONTENT_LENGTH'] = '10'
        self.environ['CONTENT_TYPE'] = ''
        self.assertRaisesRegex(webob.exc.HTTPBadRequest,
                               "content-type header required when "
                               "content-length > 0", self.app,
                               self.environ, start_response)

    def test_empty_content_length_and_type_works(self):
        self.environ['CONTENT_LENGTH'] = ''
        self.environ['CONTENT_TYPE'] = ''
        self.app(self.environ, start_response)

    def test_content_length_and_type_works(self):
        self.environ['CONTENT_LENGTH'] = '10'
        self.environ['CONTENT_TYPE'] = 'foo'
        self.app(self.environ, start_response)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Placement API schemas for setting and deleting allocations."""

import copy

from placement.schemas import common


ALLOCATION_SCHEMA = {
    "type": "object",
    "properties": {
        "allocations": {
            "type": "array",
            "minItems": 1,
            "items": {
                "type": "object",
                "properties": {
                    "resource_provider": {
                        "type": "object",
                        "properties": {
                            "uuid": {
                                "type": "string",
                                "format": "uuid"
                            }
                        },
                        "additionalProperties": False,
                        "required": ["uuid"]
                    },
                    "resources": {
                        "type": "object",
                        "minProperties": 1,
                        "patternProperties": {
                            common.RC_PATTERN: {
                                "type": "integer",
                                "minimum": 1,
                            }
                        },
                        "additionalProperties": False
                    }
                },
                "required": [
                    "resource_provider",
                    "resources"
                ],
                "additionalProperties": False
            }
        }
    },
    "required": ["allocations"],
    "additionalProperties": False
}

ALLOCATION_SCHEMA_V1_8 = copy.deepcopy(ALLOCATION_SCHEMA)
ALLOCATION_SCHEMA_V1_8['properties']['project_id'] = {'type': 'string',
                                                      'minLength': 1,
                                                      'maxLength': 255}
ALLOCATION_SCHEMA_V1_8['properties']['user_id'] = {'type': 'string',
                                                   'minLength': 1,
                                                   'maxLength': 255}
ALLOCATION_SCHEMA_V1_8['required'].extend(['project_id', 'user_id'])

# Update the allocation schema to achieve symmetry with the representation
# used when GET /allocations/{consumer_uuid} is called.
# NOTE(cdent): Explicit duplication here for sake of comprehensibility.
ALLOCATION_SCHEMA_V1_12 = {
    "type": "object",
    "properties": {
        "allocations": {
            "type": "object",
            "minProperties": 1,
            # resource provider uuid
            "patternProperties": {
                common.UUID_PATTERN: {
                    "type": "object",
                    "properties": {
                        # generation is optional
                        "generation": {
                            "type": "integer",
                        },
                        "resources": {
                            "type": "object",
                            "minProperties": 1,
                            # resource class
                            "patternProperties": {
                                common.RC_PATTERN: {
                                    "type": "integer",
                                    "minimum": 1,
                                }
                            },
                            "additionalProperties": False
                        }
                    },
                    "required": ["resources"],
                    "additionalProperties": False
                }
            },
            "additionalProperties": False
        },
        "project_id": {
            "type": "string",
            "minLength": 1,
            "maxLength": 255
        },
        "user_id": {
            "type": "string",
            "minLength": 1,
            "maxLength": 255
        }
    },
    "additionalProperties": False,
    "required": [
        "allocations",
        "project_id",
        "user_id"
    ]
}


# POST to /allocations, added in microversion 1.13, uses the
# POST_ALLOCATIONS_V1_13 schema to allow multiple allocations
# from multiple consumers in one request. It is a dict, keyed by
# consumer uuid, using the form of PUT allocations from microversion
# 1.12. In POST the allocations can be empty, so DELETABLE_ALLOCATIONS
# modifies ALLOCATION_SCHEMA_V1_12 accordingly.
DELETABLE_ALLOCATIONS = copy.deepcopy(ALLOCATION_SCHEMA_V1_12)
DELETABLE_ALLOCATIONS['properties']['allocations']['minProperties'] = 0
POST_ALLOCATIONS_V1_13 = {
    "type": "object",
    "minProperties": 1,
    "additionalProperties": False,
    "patternProperties": {
        common.UUID_PATTERN: DELETABLE_ALLOCATIONS
    }
}

# A required consumer generation was added to the top-level dict in this
# version of PUT /allocations/{consumer_uuid}. In addition, the PUT
# /allocations/{consumer_uuid}/now allows for empty allocations (indicating the
# allocations are being removed)
ALLOCATION_SCHEMA_V1_28 = copy.deepcopy(DELETABLE_ALLOCATIONS)
ALLOCATION_SCHEMA_V1_28['properties']['consumer_generation'] = {
    "type": ["integer", "null"],
    "additionalProperties": False
}
ALLOCATION_SCHEMA_V1_28['required'].append("consumer_generation")

# A required consumer generation was added to the allocations dicts in this
# version of POST /allocations
REQUIRED_GENERATION_ALLOCS_POST = copy.deepcopy(DELETABLE_ALLOCATIONS)
alloc_props = REQUIRED_GENERATION_ALLOCS_POST['properties']
alloc_props['consumer_generation'] = {
    "type": ["integer", "null"],
    "additionalProperties": False
}
REQUIRED_GENERATION_ALLOCS_POST['required'].append("consumer_generation")
POST_ALLOCATIONS_V1_28 = copy.deepcopy(POST_ALLOCATIONS_V1_13)
POST_ALLOCATIONS_V1_28["patternProperties"] = {
    common.UUID_PATTERN: REQUIRED_GENERATION_ALLOCS_POST
}

# Microversion 1.34 allows an optional mappings object which associates
# request group suffixes with lists of resource provider uuids.
mappings_schema = {
    "type": "object",
    "minProperties": 1,
    "patternProperties": {
        common.GROUP_PAT_1_33: {
            "type": "array",
            "minItems": 1,
            "items": {
                "type": "string",
                "format": "uuid"
            }
        }
    }
}
ALLOCATION_SCHEMA_V1_34 = copy.deepcopy(ALLOCATION_SCHEMA_V1_28)
ALLOCATION_SCHEMA_V1_34['properties']['mappings'] = mappings_schema
POST_ALLOCATIONS_V1_34 = copy.deepcopy(POST_ALLOCATIONS_V1_28)
POST_ALLOCATIONS_V1_34["patternProperties"] = {
    common.UUID_PATTERN: ALLOCATION_SCHEMA_V1_34
}

# A required consumer type was added to the allocations dicts in this
# version of PUT /allocations/{consumer_uuid} and POST /allocations.
ALLOCATION_SCHEMA_V1_38 = copy.deepcopy(ALLOCATION_SCHEMA_V1_34)
ALLOCATION_SCHEMA_V1_38['properties']['consumer_type'] = {
    "type": "string",
    "pattern": common.CONSUMER_TYPE_PATTERN,
    "minLength": 1,
    "maxLength": 255,
}
ALLOCATION_SCHEMA_V1_38['required'].append("consumer_type")
POST_ALLOCATIONS_V1_38 = copy.deepcopy(POST_ALLOCATIONS_V1_34)
POST_ALLOCATIONS_V1_38["patternProperties"] = {
    common.UUID_PATTERN: ALLOCATION_SCHEMA_V1_38
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the placement request log middleware."""

from unittest import mock

import testtools
import webob

from placement import requestlog


class TestRequestLog(testtools.TestCase):

    @staticmethod
    @webob.dec.wsgify
    def application(req):
        req.response.status = 200
        return req.response

    def setUp(self):
        super(TestRequestLog, self).setUp()
        self.req = webob.Request.blank('/resource_providers?name=myrp')
        self.environ = self.req.environ
        # The blank does not include remote address, so add it.
        self.environ['REMOTE_ADDR'] = '127.0.0.1'
        # nor a microversion
        self.environ['placement.microversion'] = '2.1'

    def test_get_uri(self):
        req_uri = requestlog.RequestLog._get_uri(self.environ)
        self.assertEqual('/resource_providers?name=myrp', req_uri)

    def test_get_uri_knows_prefix(self):
        self.environ['SCRIPT_NAME'] = '/placement'
        req_uri = requestlog.RequestLog._get_uri(self.environ)
        self.assertEqual('/placement/resource_providers?name=myrp', req_uri)

    @mock.patch("placement.requestlog.RequestLog.write_log")
    @mock.patch("placement.requestlog.LOG")
    def test_middleware_writes_logs(self, mocked_log, write_log):
        mocked_log.isEnabledFor.return_value = True
        start_response_mock = mock.MagicMock()
        app = requestlog.RequestLog(self.application)
        app(self.environ, start_response_mock)
        write_log.assert_called_once_with(
            self.environ, '/resource_providers?name=myrp', '200 OK', '0')

    @mock.patch("placement.requestlog.LOG")
    def test_middleware_sends_message(self, mocked_log):
        start_response_mock = mock.MagicMock()
        app = requestlog.RequestLog(self.application)
        app(self.environ, start_response_mock)
        mocked_log.debug.assert_called_once_with(
            'Starting request: %s "%s %s"', '127.0.0.1', 'GET',
            '/resource_providers?name=myrp')
        mocked_log.info.assert_called_once_with(
            '%(REMOTE_ADDR)s "%(REQUEST_METHOD)s %(REQUEST_URI)s" '
            'status: %(status)s len: %(bytes)s microversion: %(microversion)s',
            {'microversion': '2.1',
             'status': '200',
             'REQUEST_URI': '/resource_providers?name=myrp',
             'REQUEST_METHOD': 'GET',
             'REMOTE_ADDR': '127.0.0.1',
             'bytes': '0'})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Database context manager for placement database connection."""

from oslo_db.sqlalchemy import enginefacade
from oslo_log import log as logging

from placement.util import run_once

LOG = logging.getLogger(__name__)
placement_context_manager = enginefacade.transaction_context()


def _get_db_conf(conf_group):
    conf_dict = dict(conf_group.items())
    # Remove the 'sync_on_startup' conf setting, enginefacade does not use it.
    # Use pop since it might not be present in testing situations and we
    # don't want to care here.
    conf_dict.pop('sync_on_startup', None)
    return conf_dict


@run_once("TransactionFactory already started, not reconfiguring.",
          LOG.warning)
def configure(conf):
    placement_context_manager.configure(
        **_get_db_conf(conf.placement_database))


def get_placement_engine():
    return placement_context_manager.writer.get_engine()


@enginefacade.transaction_context_provider
class DbContext(object):
    """Stub class for db session handling outside of web requests."""



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_policy import opts as policy_opts
from oslo_utils.fixture import uuidsentinel

from placement import conf
from placement import direct
from placement.tests.functional import base


class TestDirect(base.TestCase):

    def setUp(self):
        super(TestDirect, self).setUp()
        self.conf = cfg.ConfigOpts()
        conf.register_opts(self.conf)
        policy_opts.set_defaults(self.conf)

    def test_direct_is_there(self):
        with direct.PlacementDirect(self.conf) as client:
            resp = client.get('/')
            self.assertTrue(resp)
            data = resp.json()
            self.assertEqual('v1.0', data['versions'][0]['id'])

    def test_get_resource_providers(self):
        with direct.PlacementDirect(self.conf) as client:
            resp = client.get('/resource_providers')
            self.assertTrue(resp)
            data = resp.json()
            self.assertEqual([], data['resource_providers'])

    def test_create_resource_provider(self):
        data = {'name': 'fake'}
        with direct.PlacementDirect(self.conf) as client:
            resp = client.post('/resource_providers', json=data)
            self.assertTrue(resp)
            resp = client.get('/resource_providers')
            self.assertTrue(resp)
            data = resp.json()
            self.assertEqual(1, len(data['resource_providers']))

    def test_json_validation_happens(self):
        data = {'name': 'fake', 'cowsay': 'moo'}
        with direct.PlacementDirect(self.conf) as client:
            resp = client.post('/resource_providers', json=data)
            self.assertFalse(resp)
            self.assertEqual(400, resp.status_code)

    def test_microversion_handling(self):
        with direct.PlacementDirect(self.conf) as client:
            # create parent
            parent_data = {'name': uuidsentinel.p_rp,
                           'uuid': uuidsentinel.p_rp}
            resp = client.post('/resource_providers', json=parent_data)
            self.assertTrue(resp, resp.text)

            # attempt to create child
            data = {'name': 'child', 'parent_provider_uuid': uuidsentinel.p_rp}
            # no microversion, 400
            resp = client.post('/resource_providers', json=data)
            self.assertFalse(resp)
            self.assertEqual(400, resp.status_code)
            # low microversion, 400
            resp = client.post('/resource_providers', json=data,
                               microversion='1.13')
            self.assertFalse(resp)
            self.assertEqual(400, resp.status_code)
            resp = client.post('/resource_providers', json=data,
                               microversion='1.14')
            self.assertTrue(resp, resp.text)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from oslo_utils.fixture import uuidsentinel as uuids
from unittest import mock

from placement import lib as placement_lib
from placement.objects import allocation_candidate as ac_obj
from placement.objects import research_context as res_ctx
from placement.objects import resource_provider as rp_obj
from placement.tests.unit.objects import base


class TestAllocationCandidatesNoDB(base.TestCase):
    @mock.patch('placement.objects.research_context._has_provider_trees',
                new=mock.Mock(return_value=True))
    def test_limit_results(self):
        # Results are limited based on their root provider uuid, not uuid.
        # For a more "real" test of this functionality, one that exercises
        # nested providers, see the 'get allocation candidates nested limit'
        # test in the 'allocation-candidates.yaml' gabbit.
        aro_in = [
            mock.Mock(
                resource_requests=[
                    mock.Mock(resource_provider=mock.Mock(
                        root_provider_uuid=uuid))
                    for uuid in (1, 0, 4, 8)]),
            mock.Mock(
                resource_requests=[
                    mock.Mock(resource_provider=mock.Mock(
                        root_provider_uuid=uuid))
                    for uuid in (4, 8, 5)]),
            mock.Mock(
                resource_requests=[
                    mock.Mock(resource_provider=mock.Mock(
                        root_provider_uuid=uuid))
                    for uuid in (1, 7, 6, 4, 8, 5)]),
        ]
        sum1 = mock.Mock(resource_provider=mock.Mock(root_provider_uuid=1))
        sum0 = mock.Mock(resource_provider=mock.Mock(root_provider_uuid=0))
        sum4 = mock.Mock(resource_provider=mock.Mock(root_provider_uuid=4))
        sum8 = mock.Mock(resource_provider=mock.Mock(root_provider_uuid=8))
        sum5 = mock.Mock(resource_provider=mock.Mock(root_provider_uuid=5))
        sum7 = mock.Mock(resource_provider=mock.Mock(root_provider_uuid=7))
        sum6 = mock.Mock(resource_provider=mock.Mock(root_provider_uuid=6))
        sum_in = [sum1, sum0, sum4, sum8, sum5, sum7, sum6]
        rw_ctx = res_ctx.RequestWideSearchContext(
            self.context, placement_lib.RequestWideParams(limit=2), True)
        aro, sum = rw_ctx.limit_results(aro_in, sum_in)
        self.assertEqual(aro_in[:2], aro)
        self.assertEqual(set([sum1, sum0, sum4, sum8, sum5]), set(sum))

    def test_check_same_subtree(self):
        # Construct a tree that look like this
        #
        #  0 -+- 00 --- 000    1 -+- 10 --- 100
        #     |                   |
        #     +- 01 -+- 010       +- 11 -+- 110
        #     |      +- 011       |      +- 111
        #     +- 02 -+- 020       +- 12 -+- 120
        #            +- 021              +- 121
        #
        parent_by_rp = {"0": None, "00": "0", "000": "00",
                        "01": "0", "010": "01", "011": "01",
                        "02": "0", "020": "02", "021": "02",
                        "1": None, "10": "1", "100": "10",
                        "11": "1", "110": "11", "111": "11",
                        "12": "1", "120": "12", "121": "12"}
        same_subtree = [
            set(["0", "00", "01"]),
            set(["01", "010"]),
            set(["02", "020", "021"]),
            set(["02", "020", "021"]),
            set(["0", "02", "010"]),
            set(["000"])
        ]

        different_subtree = [
            set(["10", "11"]),
            set(["110", "111"]),
            set(["10", "11", "110"]),
            set(["12", "120", "100"]),
            set(["0", "1"]),
        ]

        for group in same_subtree:
            self.assertTrue(
                ac_obj._check_same_subtree(group, parent_by_rp))

        for group in different_subtree:
            self.assertFalse(
                ac_obj._check_same_subtree(group, parent_by_rp))

    @mock.patch('placement.objects.research_context._has_provider_trees',
                new=mock.Mock(return_value=True))
    def _test_generate_areq_list(self, strategy, expected_candidates):
        self.conf_fixture.conf.set_override(
            "allocation_candidates_generation_strategy", strategy,
            group="placement")

        rw_ctx = res_ctx.RequestWideSearchContext(
            self.context, placement_lib.RequestWideParams(), True)
        areq_lists_by_anchor = {
            "root1": {
                "": ["r1A", "r1B",],
                "group1": ["r1g1A", "r1g1B",],
            },
            "root2": {
                "": ["r2A"],
                "group1": ["r2g1A", "r2g1B"],
            },
            "root3": {
                "": ["r3A"],
            },
        }
        generator = ac_obj._generate_areq_lists(
            rw_ctx, areq_lists_by_anchor, {"", "group1"})

        self.assertEqual(expected_candidates, list(generator))

    def test_generate_areq_lists_depth_first(self):
        # Depth-first will generate all root1 candidates first then root2,
        # root3 is ignored as it has no candidate for group1.
        expected_candidates = [
            ('r1A', 'r1g1A'),
            ('r1A', 'r1g1B'),
            ('r1B', 'r1g1A'),
            ('r1B', 'r1g1B'),
            ('r2A', 'r2g1A'),
            ('r2A', 'r2g1B'),
        ]
        self._test_generate_areq_list("depth-first", expected_candidates)

    @mock.patch('placement.objects.research_context._has_provider_trees',
                new=mock.Mock(return_value=True))
    def test_generate_areq_lists_breadth_first(self):
        # Breadth-first will take one candidate from root1 then root2 then goes
        # back to root1 etc. Root2 runs out of candidates earlier than root1 so
        # the last two candidates are both from root1. The root3 is still
        # ignored as it has no candidates for group1.
        expected_candidates = [
            ('r1A', 'r1g1A'),
            ('r2A', 'r2g1A'),
            ('r1A', 'r1g1B'),
            ('r2A', 'r2g1B'),
            ('r1B', 'r1g1A'),
            ('r1B', 'r1g1B')
        ]
        self._test_generate_areq_list("breadth-first", expected_candidates)


def _rp(rp_id, capacity=1, max_unit=None):
    return {(rp_id, "SRIOV_VF"):
            ac_obj.ProviderSummaryResource(
                resource_class="SRIOV_VF", capacity=capacity,
                used=0, max_unit=max_unit or capacity)}


def _alloc_req(group, rp_id, amount):
    return ac_obj.AllocationRequest(
        anchor_root_provider_uuid=uuids.root,
        use_same_provider=True,
        resource_requests=[
            ac_obj.AllocationRequestResource(
                resource_provider=rp_obj.ResourceProvider(
                    context=None, id=rp_id, uuid=rp_id),
                resource_class="SRIOV_VF",
                amount=amount)
        ],
        mappings={group: [rp_id]})


class TestOptimizedAllocationCandidatesNoDB(base.TestCase):
    def setUp(self):
        super().setUp()
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")

    @mock.patch('placement.objects.research_context._has_provider_trees',
                new=mock.Mock(return_value=True))
    def test_multiple_groups_usage_overlap_3_3(self):
        rw_ctx = res_ctx.RequestWideSearchContext(
            self.context, placement_lib.RequestWideParams(), True)

        # We have 3 child RPs each having a capacity of one resource
        rw_ctx.psum_res_by_rp_rc.update(_rp("RP1", capacity=1))
        rw_ctx.psum_res_by_rp_rc.update(_rp("RP2", capacity=1))
        rw_ctx.psum_res_by_rp_rc.update(_rp("RP3", capacity=1))

        G1_RP1 = _alloc_req("G1", rp_id="RP1", amount=1)
        G1_RP2 = _alloc_req("G1", rp_id="RP2", amount=1)
        G1_RP3 = _alloc_req("G1", rp_id="RP3", amount=1)

        G2_RP1 = _alloc_req("G2", rp_id="RP1", amount=1)
        G2_RP2 = _alloc_req("G2", rp_id="RP2", amount=1)
        G2_RP3 = _alloc_req("G2", rp_id="RP3", amount=1)

        G3_RP1 = _alloc_req("G3", rp_id="RP1", amount=1)
        G3_RP2 = _alloc_req("G3", rp_id="RP2", amount=1)
        G3_RP3 = _alloc_req("G3", rp_id="RP3", amount=1)

        # This algorithm starts with the possible solutions for each group
        # independent of the other groups in the same request. So here G1
        # can be fulfilled from RP1, RP2, or RP3. As well as G2, and G3.
        areq_lists_by_anchor = {
            "root": {
                "G1": [G1_RP1, G1_RP2, G1_RP3],
                "G2": [G2_RP1, G2_RP2, G2_RP3],
                "G3": [G3_RP1, G3_RP2, G3_RP3],
            },
        }

        orig_consolidate = ac_obj._consolidate_allocation_requests
        consolidate_calls = []

        def wrap_consolidate_allocation_requests(areq_list, rw_ctx):
            # we need to deep copy the call args as they are mutated during
            # the test run
            consolidate_calls.append(copy.deepcopy(areq_list))
            return orig_consolidate(areq_list, rw_ctx)

        orig_exceeds_capacity = ac_obj._exceeds_capacity
        # pairs of areq_list from the call arg and the return value of the
        # wrapped call
        exceeds_capacity_calls = []

        def warp_exceeds_capacity(rw_ctx, areq_list):
            result = orig_exceeds_capacity(rw_ctx, areq_list)
            exceeds_capacity_calls.append((areq_list, result))
            return result

        with (
            # Consolidate is not called during product generation as we
            # re-implemented exceeds_capacity to work on non consolidated
            # areq_lists
            mock.patch.object(
                ac_obj, "_consolidate_allocation_requests",
                new=mock.NonCallableMock
            ),
            # the rw_ctx exceeds_capacity working on consolidated areqs
            # are not called at all, the local _exceeds_capacity called instead
            # on all partial products, mocked below.
            mock.patch.object(
                rw_ctx, "exceeds_capacity", new=mock.NonCallableMock()
            ),
            mock.patch.object(
                ac_obj, "_exceeds_capacity",
                side_effect=warp_exceeds_capacity
            ) as mock_exceeds_capacity,
        ):
            generator = ac_obj._generate_areq_lists(
                rw_ctx, areq_lists_by_anchor, {"G1", "G2", "G3"})

            areq_lists = list(generator)

            # We don't have 27 (3^3) valid solutions just 6 (3!) as if one of
            # the Group is fulfilled from an RP then the other Groups cannot be
            # fulfilled from the same RP as each RP has 1 resource only.
            # Without the optimize_for_wide_provider_trees = true the
            # _generate_areq_lists call would generate all the 27 possible
            # product and then later processing would filter out the invalid,
            # overlapping ones.
            self.assertEqual(6, len(areq_lists))
            self.assertEqual([
                (G1_RP1, G2_RP2, G3_RP3),
                (G1_RP1, G2_RP3, G3_RP2),
                (G1_RP2, G2_RP1, G3_RP3),
                (G1_RP2, G2_RP3, G3_RP1),
                (G1_RP3, G2_RP1, G3_RP2),
                (G1_RP3, G2_RP2, G3_RP1),],
                areq_lists)

        # areq_list, result pairs where areq_list is a partial product
        # and the result is the expected return value of _exceeds_capacity
        expected_exceeds_capacity_calls = [
            ((G1_RP1,), False),
            ((G1_RP1, G2_RP1), True),
            # This is the pruning. The algo did not try to generate
            # G1_RP1, G2_RP1, G3_RPx for all three possible x RPs as it knows
            # that if the prefix is invalid then all the product with that
            # prefix is also invalid.
            ((G1_RP1, G2_RP2), False),
            ((G1_RP1, G2_RP2, G3_RP1), True),
            # This is another optimization (compared to an index odometer
            # based product algo) that the recursive call reuses the already
            # calculated and checked valid prefix of G1_RP1, G2_RP2 and don't
            # need to re-create it to try G3_RP2 with it.
            ((G1_RP1, G2_RP2, G3_RP2), True),
            ((G1_RP1, G2_RP2, G3_RP3), False),  # this is a valid product
            # simple backtrack as we run out of possibilities on level 3
            ((G1_RP1, G2_RP3), False),
            ((G1_RP1, G2_RP3, G3_RP1), True),
            ((G1_RP1, G2_RP3, G3_RP2), False),  # this is a valid product
            ((G1_RP1, G2_RP3, G3_RP3), True),
            # double backtrack as we run out the possibilities on level 2 and
            # level 3
            ((G1_RP2,), False),
            ((G1_RP2, G2_RP1), False),
            ((G1_RP2, G2_RP1, G3_RP1), True),
            ((G1_RP2, G2_RP1, G3_RP2), True),
            ((G1_RP2, G2_RP1, G3_RP3), False),  # this is a valid product
            ((G1_RP2, G2_RP2), True),  # suffixes are pruned
            ((G1_RP2, G2_RP3), False),
            ((G1_RP2, G2_RP3, G3_RP1), False),  # this is a valid product
            ((G1_RP2, G2_RP3, G3_RP2), True),
            ((G1_RP2, G2_RP3, G3_RP3), True),  # double backtrack
            ((G1_RP3,), False),
            ((G1_RP3, G2_RP1), False),
            ((G1_RP3, G2_RP1, G3_RP1), True),
            ((G1_RP3, G2_RP1, G3_RP2), False),  # this is a valid product
            ((G1_RP3, G2_RP1, G3_RP3), True),   # backtrack
            ((G1_RP3, G2_RP2), False),
            ((G1_RP3, G2_RP2, G3_RP1), False),  # this is a valid product
            ((G1_RP3, G2_RP2, G3_RP2), True),
            ((G1_RP3, G2_RP2, G3_RP3), True),  # backtrack
            ((G1_RP3, G2_RP3), True),  # suffixes are pruned
            # backtrack all the way as we finished with the level 1
            # possibilities as well
        ]

        for i, p in enumerate(
            zip(expected_exceeds_capacity_calls, exceeds_capacity_calls)
        ):
            expected, actual = p
            self.assertEqual(
                expected, actual, "Call index %d does not match" % i)

        # Why 30 is the correct answer?
        # I don't have an easy way to prove that mathematically. It was
        # easier to list them all out above so they can be reviewed
        #
        # With this 3 by 3 example the number of checks are higher when
        # optimization is enabled than without it. But if you increase the
        # size of the product then the number of pruned space grows fast
        # as well as the size of the product space. On an 8 by 8 example
        # this is already a must-have optimization to run the checks in
        # less than a minute. See the included functional tests for the
        # scale results.
        self.assertEqual(30, len(mock_exceeds_capacity.mock_calls))
        self.assertEqual(30, len(expected_exceeds_capacity_calls))


class TestExceedsCapacityNoDB(base.TestCase):

    def setUp(self):
        super().setUp()

        patcher = mock.patch(
            'placement.objects.research_context._has_provider_trees',
            new=mock.Mock(return_value=True))
        self.addCleanup(patcher.stop)
        patcher.start()

        self.rw_ctx = res_ctx.RequestWideSearchContext(
            self.context, placement_lib.RequestWideParams(), True)

        self.rw_ctx.psum_res_by_rp_rc.update(
            _rp("RP1", capacity=3, max_unit=1))
        self.rw_ctx.psum_res_by_rp_rc.update(
            _rp("RP2", capacity=2, max_unit=2))
        self.rw_ctx.psum_res_by_rp_rc.update(
            _rp("RP3", capacity=1, max_unit=1))

    def test_not_exceeds(self):
        self.assertFalse(
            ac_obj._exceeds_capacity(
                self.rw_ctx, (_alloc_req("G1", rp_id="RP1", amount=1),)))

        self.assertFalse(
            ac_obj._exceeds_capacity(
                self.rw_ctx, (
                    _alloc_req("G1", rp_id="RP2", amount=1),
                    _alloc_req("G2", rp_id="RP2", amount=1),
                )))

        self.assertFalse(
            ac_obj._exceeds_capacity(
                self.rw_ctx, (
                    _alloc_req("G1", rp_id="RP3", amount=1),
                    _alloc_req("G2", rp_id="RP1", amount=1),
                )))

    def test_exceeds_capacity(self):
        self.assertTrue(
            ac_obj._exceeds_capacity(
                self.rw_ctx, (
                    _alloc_req("G1", rp_id="RP3", amount=1),
                    _alloc_req("G2", rp_id="RP3", amount=1),
                )))

    def test_exceeds_max_unit(self):
        self.assertTrue(
            ac_obj._exceeds_capacity(self.rw_ctx, (
                _alloc_req("G1", rp_id="RP1", amount=1),
                _alloc_req("G2", rp_id="RP1", amount=1)
            )))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for code in the aggregate handler that gabbi isn't covering."""

from unittest import mock

import webob

from placement import context
from placement import exception
from placement.handlers import aggregate
from placement.objects import resource_provider
from placement.tests.unit import base


class TestAggregateHandlerErrors(base.ContextTestCase):
    """Tests that make sure errors hard to trigger by gabbi result in expected
    exceptions.
    """

    def test_concurrent_exception_causes_409(self):
        fake_context = context.RequestContext(
            user_id='fake', project_id='fake')
        rp = resource_provider.ResourceProvider(fake_context)
        expected_message = ('Update conflict: Another thread concurrently '
                            'updated the data')
        with mock.patch("placement.objects.resource_provider._set_aggregates",
                        side_effect=exception.ConcurrentUpdateDetected):
            exc = self.assertRaises(webob.exc.HTTPConflict,
                                    aggregate._set_aggregates, rp, [])
        self.assertIn(expected_message, str(exc))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for database migrations. There are "opportunistic" tests for sqlite in
memory, mysql and postgresql in here, which allows testing against these
databases in a properly configured unit test environment.

For the opportunistic testing you need to set up a db named 'openstack_citest'
with user 'openstack_citest' and password 'openstack_citest' on localhost. This
can be accomplished by running the `test-setup.sh` script in the `tools`
subdirectory. The test will then use that DB and username/password combo to run
the tests.
"""

from unittest import mock

from alembic import script
from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import test_migrations
from oslo_db.sqlalchemy import utils as db_utils
from oslo_log import log as logging
from oslo_utils.fixture import uuidsentinel as uuids
from sqlalchemy import inspect

from placement.db.sqlalchemy import migration
from placement.db.sqlalchemy import models
from placement import db_api
from placement.tests.functional import base


LOG = logging.getLogger(__name__)


class WalkVersionsMixin(object):
    def _walk_versions(self):
        """Determine latest version script from the repo, then upgrade from 1
        through to the latest, with no data in the databases. This just checks
        that the schema itself upgrades successfully.
        """

        # Place the database under version control
        script_directory = script.ScriptDirectory.from_config(self.config)
        self.assertIsNone(self.migration_api.version(self.config))
        versions = [ver for ver in script_directory.walk_revisions()]
        for version in reversed(versions):
            self._migrate_up(version.revision, with_data=True)

    def _migrate_up(self, version, with_data=False):
        """Migrate up to a new version of the db.

        We allow for data insertion and post checks at every
        migration version with special _pre_upgrade_### and
        _check_### functions in the main test.
        """
        # NOTE(sdague): try block is here because it's impossible to debug
        # where a failed data migration happens otherwise
        try:
            if with_data:
                data = None
                pre_upgrade = getattr(
                    self, "_pre_upgrade_%s" % version, None)
                if pre_upgrade:
                    data = pre_upgrade(self.engine)

            self.migration_api.upgrade(version, config=self.config)
            self.assertEqual(version, self.migration_api.version(self.config))
            if with_data:
                check = getattr(self, "_check_%s" % version, None)
                if check:
                    check(self.engine, data)
        except Exception:
            LOG.error("Failed to migrate to version %(version)s on engine "
                      "%(engine)s",
                      {'version': version, 'engine': self.engine})
            raise


class TestWalkVersions(base.NoDBTestCase, WalkVersionsMixin):
    def setUp(self):
        super(TestWalkVersions, self).setUp()
        self.migration_api = mock.MagicMock()
        self.engine = mock.MagicMock()
        self.config = mock.MagicMock()
        self.versions = [mock.Mock(revision='2b2'), mock.Mock(revision='1a1')]

    def test_migrate_up(self):
        self.migration_api.version.return_value = 'dsa123'
        self._migrate_up('dsa123')
        self.migration_api.upgrade.assert_called_with('dsa123',
                                                      config=self.config)
        self.migration_api.version.assert_called_with(self.config)

    def test_migrate_up_with_data(self):
        test_value = {"a": 1, "b": 2}
        self.migration_api.version.return_value = '141'
        self._pre_upgrade_141 = mock.MagicMock()
        self._pre_upgrade_141.return_value = test_value
        self._check_141 = mock.MagicMock()
        self._migrate_up('141', True)
        self._pre_upgrade_141.assert_called_with(self.engine)
        self._check_141.assert_called_with(self.engine, test_value)

    @mock.patch.object(script, 'ScriptDirectory')
    @mock.patch.object(WalkVersionsMixin, '_migrate_up')
    def test_walk_versions_all_default(self, _migrate_up, script_directory):
        fc = script_directory.from_config()
        fc.walk_revisions.return_value = self.versions
        self.migration_api.version.return_value = None
        self._walk_versions()
        self.migration_api.version.assert_called_with(self.config)
        upgraded = [mock.call(v.revision,
                    with_data=True) for v in reversed(self.versions)]
        self.assertEqual(self._migrate_up.call_args_list, upgraded)

    @mock.patch.object(script, 'ScriptDirectory')
    @mock.patch.object(WalkVersionsMixin, '_migrate_up')
    def test_walk_versions_all_false(self, _migrate_up, script_directory):
        fc = script_directory.from_config()
        fc.walk_revisions.return_value = self.versions
        self.migration_api.version.return_value = None
        self._walk_versions()
        upgraded = [mock.call(v.revision,
                    with_data=True) for v in reversed(self.versions)]
        self.assertEqual(upgraded, self._migrate_up.call_args_list)


class MigrationCheckersMixin(object):
    def setUp(self):
        super(MigrationCheckersMixin, self).setUp()
        self.engine = db_api.placement_context_manager.writer.get_engine()
        self.config = migration._alembic_config()
        self.migration_api = migration

    def test_walk_versions(self):
        self._walk_versions()

#    # Leaving this here as a sort of template for when we do migration tests.
#    def _check_fb3f10dd262e(self, engine, data):
#        nodes_tbl = db_utils.get_table(engine, 'nodes')
#        col_names = [column.name for column in nodes_tbl.c]
#        self.assertIn('fault', col_names)
#        self.assertIsInstance(nodes_tbl.c.fault.type,
#                              sqlalchemy.types.String)

    def test_upgrade_and_version(self):
        self.migration_api.upgrade('head')
        self.assertIsNotNone(self.migration_api.version())

    def test_upgrade_twice(self):
        # Start with the empty version
        self.migration_api.upgrade('base')
        v1 = self.migration_api.version()
        # Now upgrade to head
        self.migration_api.upgrade('head')
        v2 = self.migration_api.version()
        self.assertNotEqual(v1, v2)

    def test_block_on_null_root_provider_id(self):
        """Upgrades the schema to b4ed3a175331 (initial), injects a resource
        provider with no root provider and then tries to upgrade to head which
        should fail on the 611cd6dffd7b blocker migration.
        """
        # Upgrade to populate the schema.
        self.migration_api.upgrade('b4ed3a175331')
        # Now insert a resource provider with no root.
        rps = db_utils.get_table(self.engine, 'resource_providers')
        ins_stmt = rps.insert().values(
            name='fake-rp-name',
            uuid=uuids.rp_uuid,
        )
        with self.engine.connect() as conn, conn.begin():
            rp_id = conn.execute(ins_stmt).inserted_primary_key[0]
        # Now run the blocker migration and it should raise an error.
        ex = self.assertRaises(  # noqa H202
            Exception, self.migration_api.upgrade, '611cd6dffd7b')
        # Make sure it's the error we expect.
        self.assertIn('There is at least one resource provider table '
                      'record which is missing its root provider id.',
                      str(ex))
        # Now update the resource provider with a root_provider_id.
        update_stmt = rps.update().values(
            root_provider_id=rp_id,
        ).where(rps.c.id == rp_id)
        with self.engine.connect() as conn, conn.begin():
            conn.execute(update_stmt)
        # Re-run the upgrade and it should be OK.
        self.migration_api.upgrade('611cd6dffd7b')

    def test_block_on_missing_consumer(self):
        """Upgrades the schema to b4ed3a175331 (initial), injects an allocation
        without a corresponding consumer record and then tries to upgrade to
        head which should fail on the b5c396305c25 blocker migration.
        """
        # Upgrade to populate the schema.
        self.migration_api.upgrade('b4ed3a175331')
        # Now insert a resource provider to build off
        rps = db_utils.get_table(self.engine, 'resource_providers')
        ins_stmt = rps.insert().values(
            name='fake-rp-name',
            uuid=uuids.rp_uuid,
            root_provider_id=1,
        )
        with self.engine.connect() as conn, conn.begin():
            rp_id = conn.execute(ins_stmt).inserted_primary_key[0]
        # Now insert an allocation
        allocations = db_utils.get_table(self.engine, 'allocations')
        ins_stmt = allocations.insert().values(
            resource_provider_id=rp_id,
            resource_class_id=1,
            used=5,
            consumer_id=uuids.consumer1,
        )
        with self.engine.connect() as conn, conn.begin():
            conn.execute(ins_stmt).inserted_primary_key[0]
        # Now run the blocker migration and it should raise an error.
        ex = self.assertRaises(  # noqa H202
            Exception, self.migration_api.upgrade, 'b5c396305c25')
        # Make sure it's the error we expect.
        self.assertIn('There is at least one allocation record which is '
                      'missing a consumer record.',
                      str(ex))
        # Add a (faked) consumer record and try again
        consumers = db_utils.get_table(self.engine, 'consumers')
        ins_stmt = consumers.insert().values(
            uuid=uuids.consumer1,
            project_id=1,
            user_id=1,
        )
        with self.engine.connect() as conn, conn.begin():
            conn.execute(ins_stmt).inserted_primary_key[0]
        self.migration_api.upgrade('b5c396305c25')

    def test_consumer_types_422ece571366(self):
        # Upgrade to populate the schema.
        self.migration_api.upgrade('422ece571366')
        insp = inspect(self.engine)
        # Test creation of consumer_types table
        con = db_utils.get_table(self.engine, 'consumer_types')
        col_names = [column.name for column in con.c]
        self.assertIn('created_at', col_names)
        self.assertIn('updated_at', col_names)
        self.assertIn('id', col_names)
        self.assertIn('name', col_names)
        # check constraints
        pkey = insp.get_pk_constraint("consumer_types")
        self.assertEqual(['id'], pkey['constrained_columns'])
        ukey = insp.get_unique_constraints("consumer_types")
        self.assertEqual('uniq_consumer_types0name', ukey[0]['name'])

    def test_consumer_type_id_column_422ece571366(self):
        # Upgrade to populate the schema.
        self.migration_api.upgrade('422ece571366')
        insp = inspect(self.engine)
        # Test creation of consumer_types table
        consumers = db_utils.get_table(self.engine, 'consumers')
        col_names = [column.name for column in consumers.c]
        self.assertIn('consumer_type_id', col_names)
        # Check index and constraints
        fkey = insp.get_foreign_keys("consumers")
        self.assertEqual(['consumer_type_id'], fkey[0]['constrained_columns'])
        ind = insp.get_indexes('consumers')
        names = [r['name'] for r in ind]
        self.assertIn('consumers_consumer_type_id_idx', names)


class PlacementOpportunisticFixture(object):
    def get_enginefacade(self):
        return db_api.placement_context_manager


class SQLiteOpportunisticFixture(
        PlacementOpportunisticFixture, test_fixtures.OpportunisticDbFixture):
    pass


class MySQLOpportunisticFixture(
        PlacementOpportunisticFixture,
        test_fixtures.MySQLOpportunisticFixture):
    pass


class PostgresqlOpportunisticFixture(
        PlacementOpportunisticFixture,
        test_fixtures.PostgresqlOpportunisticFixture):
    pass


class TestMigrationsSQLite(MigrationCheckersMixin,
                           WalkVersionsMixin,
                           test_fixtures.OpportunisticDBTestMixin,
                           base.NoDBTestCase):
    FIXTURE = SQLiteOpportunisticFixture


class TestMigrationsMySQL(MigrationCheckersMixin,
                          WalkVersionsMixin,
                          test_fixtures.OpportunisticDBTestMixin,
                          base.NoDBTestCase):
    FIXTURE = MySQLOpportunisticFixture


class TestMigrationsPostgresql(MigrationCheckersMixin,
                               WalkVersionsMixin,
                               test_fixtures.OpportunisticDBTestMixin,
                               base.NoDBTestCase):
    FIXTURE = PostgresqlOpportunisticFixture


class _TestModelsMigrations(test_migrations.ModelsMigrationsSync):
    def get_metadata(self):
        return models.BASE.metadata

    def get_engine(self):
        return db_api.get_placement_engine()

    def db_sync(self, engine):
        migration.upgrade('head')


class ModelsMigrationsSyncSqlite(_TestModelsMigrations,
                                 test_fixtures.OpportunisticDBTestMixin,
                                 base.NoDBTestCase):
    FIXTURE = SQLiteOpportunisticFixture


class ModelsMigrationsSyncMysql(_TestModelsMigrations,
                                test_fixtures.OpportunisticDBTestMixin,
                                base.NoDBTestCase):
    FIXTURE = MySQLOpportunisticFixture


class ModelsMigrationsSyncPostgresql(_TestModelsMigrations,
                                     test_fixtures.OpportunisticDBTestMixin,
                                     base.NoDBTestCase):
    FIXTURE = PostgresqlOpportunisticFixture



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from oslo_db import api as oslo_db_api
from oslo_log import log as logging
import sqlalchemy as sa
from sqlalchemy import sql

from placement.db.sqlalchemy import models
from placement import db_api
from placement import exception
from placement.objects import consumer as consumer_obj
from placement.objects import project as project_obj
from placement.objects import resource_provider as rp_obj
from placement.objects import user as user_obj


_ALLOC_TBL = models.Allocation.__table__
_CONSUMER_TBL = models.Consumer.__table__
_INV_TBL = models.Inventory.__table__
_PROJECT_TBL = models.Project.__table__
_RP_TBL = models.ResourceProvider.__table__
_USER_TBL = models.User.__table__

LOG = logging.getLogger(__name__)


class Allocation(object):

    def __init__(self, id=None, resource_provider=None, consumer=None,
                 resource_class=None, used=0, updated_at=None,
                 created_at=None):
        self.id = id
        self.resource_provider = resource_provider
        self.resource_class = resource_class
        self.consumer = consumer
        self.used = used
        self.updated_at = updated_at
        self.created_at = created_at


@db_api.placement_context_manager.writer
def _delete_allocations_for_consumer(ctx, consumer_id):
    """Deletes any existing allocations that correspond to the allocations to
    be written. This is wrapped in a transaction, so if the write subsequently
    fails, the deletion will also be rolled back.
    """
    del_sql = _ALLOC_TBL.delete().where(
        _ALLOC_TBL.c.consumer_id == consumer_id)
    ctx.session.execute(del_sql)


@db_api.placement_context_manager.writer
def _delete_allocations_by_ids(ctx, alloc_ids):
    """Deletes allocations having an internal id value in the set of supplied
    IDs
    """
    del_sql = _ALLOC_TBL.delete().where(_ALLOC_TBL.c.id.in_(alloc_ids))
    ctx.session.execute(del_sql)


def _provider_usage_query(provider_ids, rc_ids):
    """Generate a query fragment to grab provider usage per class."""
    usage = sa.select(
        _ALLOC_TBL.c.resource_provider_id,
        _ALLOC_TBL.c.resource_class_id,
        sql.func.sum(_ALLOC_TBL.c.used).label('used'),
    )
    usage = usage.where(
        sa.and_(_ALLOC_TBL.c.resource_class_id.in_(rc_ids),
                _ALLOC_TBL.c.resource_provider_id.in_(provider_ids)))
    usage = usage.group_by(_ALLOC_TBL.c.resource_provider_id,
                           _ALLOC_TBL.c.resource_class_id)
    return usage


def _provider_usage_summary(ctx, provider_ids, rcs):
    """Generate a summary dict of usages for a given list of providers.

    Returns a dict of {rp_id: {rc1: n, rc2: m, ...}}

    :param provider_ids: List of resource provider ids
    :param rcs: List of resource class names
    """
    usage = _provider_usage_query(provider_ids,
                                  [ctx.rc_cache.id_from_string(rc)
                                   for rc in rcs])

    records = ctx.session.execute(usage)
    summary = collections.defaultdict(dict)
    for record in records:
        rp_summary = summary[record.resource_provider_id]
        rp_summary[record.resource_class_id] = record.used
    return summary


def _check_capacity_exceeded(ctx, allocs, prev_usage):
    """Checks to see if the supplied allocation records would result in any of
    the inventories involved having their capacity exceeded.

    Raises an InvalidAllocationCapacityExceeded exception if any inventory
    would be exhausted by the allocation. Raises an
    InvalidAllocationConstraintsViolated exception if any of the `step_size`,
    `min_unit` or `max_unit` constraints in an inventory will be violated
    by any one of the allocations.

    If no inventories would be exceeded or violated by the allocations, the
    function returns a list of `ResourceProvider` objects that contain the
    generation at the time of the check.

    :param ctx: `placement.context.RequestContext` that has an oslo_db
                Session
    :param allocs: List of `Allocation` objects to check
    :param prev_usage: Dict of providers' previous usage counts before this
                       operation
    """
    # The SQL generated below looks like this:
    # SELECT
    #   rp.id,
    #   rp.uuid,
    #   rp.generation,
    #   inv.resource_class_id,
    #   inv.total,
    #   inv.reserved,
    #   inv.allocation_ratio,
    #   allocs.used
    # FROM resource_providers AS rp
    # JOIN inventories AS i1
    # ON rp.id = i1.resource_provider_id
    # LEFT JOIN (
    #    SELECT resource_provider_id, resource_class_id, SUM(used) AS used
    #    FROM allocations
    #    WHERE resource_class_id IN ($RESOURCE_CLASSES)
    #    AND resource_provider_id IN ($RESOURCE_PROVIDERS)
    #    GROUP BY resource_provider_id, resource_class_id
    # ) AS allocs
    # ON inv.resource_provider_id = allocs.resource_provider_id
    # AND inv.resource_class_id = allocs.resource_class_id
    # WHERE rp.id IN ($RESOURCE_PROVIDERS)
    # AND inv.resource_class_id IN ($RESOURCE_CLASSES)
    #
    # We then take the results of the above and determine if any of the
    # inventory will have its capacity exceeded.
    rc_ids = set([ctx.rc_cache.id_from_string(a.resource_class)
                  for a in allocs])
    provider_uuids = set([a.resource_provider.uuid for a in allocs])
    provider_ids = set([a.resource_provider.id for a in allocs])
    usage = _provider_usage_query(provider_ids, rc_ids).subquery(name='usage')

    inv_join = sql.join(
        _RP_TBL, _INV_TBL,
        sql.and_(_RP_TBL.c.id == _INV_TBL.c.resource_provider_id,
                 _INV_TBL.c.resource_class_id.in_(rc_ids)))
    primary_join = sql.outerjoin(
        inv_join, usage,
        sql.and_(
            _INV_TBL.c.resource_provider_id == usage.c.resource_provider_id,
            _INV_TBL.c.resource_class_id == usage.c.resource_class_id)
    )

    sel = sa.select(
        _RP_TBL.c.id.label('resource_provider_id'),
        _RP_TBL.c.uuid,
        _RP_TBL.c.generation,
        _INV_TBL.c.resource_class_id,
        _INV_TBL.c.total,
        _INV_TBL.c.reserved,
        _INV_TBL.c.allocation_ratio,
        _INV_TBL.c.min_unit,
        _INV_TBL.c.max_unit,
        _INV_TBL.c.step_size,
        usage.c.used,
    ).select_from(primary_join)
    sel = sel.where(
        sa.and_(_RP_TBL.c.id.in_(provider_ids),
                _INV_TBL.c.resource_class_id.in_(rc_ids)))
    records = ctx.session.execute(sel)
    # Create a map keyed by (rp_uuid, res_class) for the records in the DB
    usage_map = {}
    provs_with_inv = set()
    for record in records:
        map_key = (record.uuid, record.resource_class_id)
        if map_key in usage_map:
            raise KeyError("%s already in usage_map, bad query" % str(map_key))
        usage_map[map_key] = record
        provs_with_inv.add(record.uuid)
    # Ensure that all providers have existing inventory
    missing_provs = provider_uuids - provs_with_inv
    if missing_provs:
        class_str = ', '.join([ctx.rc_cache.string_from_id(rc_id)
                               for rc_id in rc_ids])
        provider_str = ', '.join(missing_provs)
        raise exception.InvalidInventory(
            resource_class=class_str, resource_provider=provider_str)

    res_providers = {}
    rp_resource_class_sum = collections.defaultdict(
        lambda: collections.defaultdict(int))
    for alloc in allocs:
        rc_id = ctx.rc_cache.id_from_string(alloc.resource_class)
        rp_uuid = alloc.resource_provider.uuid
        if rp_uuid not in res_providers:
            res_providers[rp_uuid] = alloc.resource_provider
        amount_needed = alloc.used
        rp_resource_class_sum[rp_uuid][rc_id] += amount_needed
        # No use checking usage if we're not asking for anything
        if amount_needed == 0:
            continue
        key = (rp_uuid, rc_id)
        try:
            usage = usage_map[key]
        except KeyError:
            # The resource class at rc_id is not in the usage map.
            raise exception.InvalidInventory(
                resource_class=alloc.resource_class,
                resource_provider=rp_uuid)
        allocation_ratio = usage.allocation_ratio
        min_unit = usage.min_unit
        max_unit = usage.max_unit
        step_size = usage.step_size

        # check min_unit, max_unit, step_size
        if (amount_needed < min_unit or amount_needed > max_unit or
                amount_needed % step_size != 0):
            LOG.warning(
                "Allocation for %(rc)s on resource provider %(rp)s "
                "violates min_unit, max_unit, or step_size. "
                "Requested: %(requested)s, min_unit: %(min_unit)s, "
                "max_unit: %(max_unit)s, step_size: %(step_size)s",
                {'rc': alloc.resource_class,
                 'rp': rp_uuid,
                 'requested': amount_needed,
                 'min_unit': min_unit,
                 'max_unit': max_unit,
                 'step_size': step_size})
            raise exception.InvalidAllocationConstraintsViolated(
                resource_class=alloc.resource_class,
                resource_provider=rp_uuid)

        # usage.used can be returned as None
        used = usage.used or 0
        capacity = (usage.total - usage.reserved) * allocation_ratio
        amount_needed_all = rp_resource_class_sum[rp_uuid][rc_id]
        if (capacity < (used + amount_needed) or
                capacity < (used + amount_needed_all)):
            # If we get here, this RP is over capacity. That's a problem,
            # except we allow it *if* the current set of allocations is
            # making the situation better (i.e. resulting in less or the same
            # amount of over-capacity).

            try:
                prev_used = prev_usage[alloc.resource_provider.id][rc_id]
            except KeyError:
                prev_used = None
            params = {
                'rc': alloc.resource_class,
                'rp': rp_uuid,
                'needed': amount_needed,
                'needed_all': used + amount_needed_all,
                'used': used,
                'prev_used': prev_used,
                'cap': capacity,
            }

            if prev_used is not None and (
                    used + amount_needed_all <= prev_used):
                # The situation is no worse than it was, so no reason to
                # refuse the change.
                LOG.warning(
                    "Over capacity for %(rc)s on resource provider "
                    "%(rp)s but allocation of %(needed)s reduces "
                    "existing overage from %(prev_used)s to %(needed_all)s",
                    params)
            else:
                # The situation is worse than it was, so refuse.
                LOG.warning(
                    "Over capacity for %(rc)s on resource provider %(rp)s. "
                    "Needed: %(needed)s, Used: %(used)s, Capacity: %(cap)s",
                    params)
                raise exception.InvalidAllocationCapacityExceeded(
                    resource_class=alloc.resource_class,
                    resource_provider=rp_uuid)
    return res_providers


@db_api.placement_context_manager.reader
def _get_allocations_by_provider_id(ctx, rp_id):
    allocs = sa.alias(_ALLOC_TBL, name="a")
    consumers = sa.alias(_CONSUMER_TBL, name="c")
    projects = sa.alias(_PROJECT_TBL, name="p")
    users = sa.alias(_USER_TBL, name="u")

    # TODO(jaypipes): change this join to be on ID not UUID
    consumers_join = sa.join(
        allocs, consumers, allocs.c.consumer_id == consumers.c.uuid)
    projects_join = sa.join(
        consumers_join, projects, consumers.c.project_id == projects.c.id)
    users_join = sa.join(
        projects_join, users, consumers.c.user_id == users.c.id)

    sel = sa.select(
        allocs.c.id,
        allocs.c.resource_class_id,
        allocs.c.used,
        allocs.c.updated_at,
        allocs.c.created_at,
        consumers.c.id.label("consumer_id"),
        consumers.c.generation.label("consumer_generation"),
        consumers.c.uuid.label("consumer_uuid"),
        projects.c.id.label("project_id"),
        projects.c.external_id.label("project_external_id"),
        users.c.id.label("user_id"),
        users.c.external_id.label("user_external_id"),
    ).select_from(users_join)
    sel = sel.where(allocs.c.resource_provider_id == rp_id)

    return [dict(r._mapping) for r in ctx.session.execute(sel)]


@db_api.placement_context_manager.reader
def _get_allocations_by_consumer_uuid(ctx, consumer_uuid):
    allocs = sa.alias(_ALLOC_TBL, name="a")
    rp = sa.alias(_RP_TBL, name="rp")
    consumer = sa.alias(_CONSUMER_TBL, name="c")
    project = sa.alias(_PROJECT_TBL, name="p")
    user = sa.alias(_USER_TBL, name="u")

    # Build up the joins of the five tables we need to interact with.
    rp_join = sa.join(allocs, rp, allocs.c.resource_provider_id == rp.c.id)
    consumer_join = sa.join(
        rp_join, consumer, allocs.c.consumer_id == consumer.c.uuid)
    project_join = sa.join(
        consumer_join, project, consumer.c.project_id == project.c.id)
    user_join = sa.join(
        project_join, user, consumer.c.user_id == user.c.id)

    sel = sa.select(
        allocs.c.id,
        allocs.c.resource_provider_id,
        rp.c.name.label("resource_provider_name"),
        rp.c.uuid.label("resource_provider_uuid"),
        rp.c.generation.label("resource_provider_generation"),
        allocs.c.resource_class_id,
        allocs.c.used,
        consumer.c.id.label("consumer_id"),
        consumer.c.generation.label("consumer_generation"),
        consumer.c.consumer_type_id,
        consumer.c.uuid.label("consumer_uuid"),
        project.c.id.label("project_id"),
        project.c.external_id.label("project_external_id"),
        user.c.id.label("user_id"),
        user.c.external_id.label("user_external_id"),
        allocs.c.created_at,
        allocs.c.updated_at,
    ).select_from(user_join)
    sel = sel.where(allocs.c.consumer_id == consumer_uuid)

    return [dict(r._mapping) for r in ctx.session.execute(sel)]


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
@db_api.placement_context_manager.writer
def _set_allocations(context, allocs):
    """Write a set of allocations.

    We must check that there is capacity for each allocation.
    If there is not we roll back the entire set.

    :raises `exception.ResourceClassNotFound` if any resource class in any
            allocation in allocs cannot be found in either the DB.
    :raises `exception.InvalidAllocationCapacityExceeded` if any inventory
            would be exhausted by the allocation.
    :raises `InvalidAllocationConstraintsViolated` if any of the
            `step_size`, `min_unit` or `max_unit` constraints in an
            inventory will be violated by any one of the allocations.
    :raises `ConcurrentUpdateDetected` if a generation for a resource
            provider or consumer failed its increment check.
    """
    # First delete any existing allocations for any consumers. This
    # provides a clean slate for the consumers mentioned in the list of
    # allocations being manipulated. Before we do, grab a snapshot of the
    # current usage for any of the affected providers and classes.
    consumer_ids = set(alloc.consumer.uuid for alloc in allocs)
    prev_usage = _provider_usage_summary(
        context,
        [alloc.resource_provider.id for alloc in allocs],
        [alloc.resource_class for alloc in allocs])
    for consumer_id in consumer_ids:
        _delete_allocations_for_consumer(context, consumer_id)

    # Before writing any allocation records, we check that the submitted
    # allocations do not cause any inventory capacity to be exceeded for
    # any resource provider and resource class involved in the allocation
    # transaction. _check_capacity_exceeded() raises an exception if any
    # inventory capacity is exceeded. If capacity is not exceeded, the
    # function returns a list of ResourceProvider objects containing the
    # generation of the resource provider at the time of the check. These
    # objects are used at the end of the allocation transaction as a guard
    # against concurrent updates.
    #
    # Don't check capacity when alloc.used is zero. Zero is not a valid
    # amount when making an allocation (the minimum consumption of a
    # resource is one) but is used in this method to indicate a need for
    # removal. Providing 0 is controlled at the HTTP API layer where PUT
    # /allocations does not allow empty allocations. When POST /allocations
    # is implemented it will for the special case of atomically setting and
    # removing different allocations in the same request.
    # _check_capacity_exceeded will raise a ResourceClassNotFound # if any
    # allocation is using a resource class that does not exist.
    visited_consumers = {}
    visited_rps = _check_capacity_exceeded(context, allocs, prev_usage)
    for alloc in allocs:
        if alloc.consumer.id not in visited_consumers:
            visited_consumers[alloc.consumer.id] = alloc.consumer

        # If alloc.used is set to zero that is a signal that we don't want
        # to (re-)create any allocations for this resource class.
        # _delete_current_allocs has already wiped out allocations so just
        # continue
        if alloc.used == 0:
            continue
        consumer_id = alloc.consumer.uuid
        rp = alloc.resource_provider
        rc_id = context.rc_cache.id_from_string(alloc.resource_class)
        ins_stmt = _ALLOC_TBL.insert().values(
            resource_provider_id=rp.id,
            resource_class_id=rc_id,
            consumer_id=consumer_id,
            used=alloc.used)
        res = context.session.execute(ins_stmt)
        alloc.id = res.lastrowid

    # Generation checking happens here. If the inventory for this resource
    # provider changed out from under us, this will raise a
    # ConcurrentUpdateDetected which can be caught by the caller to choose
    # to try again. It will also rollback the transaction so that these
    # changes always happen atomically.
    for rp in visited_rps.values():
        rp.increment_generation()
    for consumer in visited_consumers.values():
        consumer.increment_generation()
    # If any consumers involved in this transaction ended up having no
    # allocations, delete the consumer records. Exclude consumers that had
    # *some resource* in the allocation list with a total > 0 since clearly
    # those consumers have allocations...
    cons_with_allocs = set(a.consumer.uuid for a in allocs if a.used > 0)
    all_cons = set(c.uuid for c in visited_consumers.values())
    consumers_to_check = all_cons - cons_with_allocs
    consumer_obj.delete_consumers_if_no_allocations(
        context, consumers_to_check)


def get_all_by_resource_provider(context, rp):
    db_allocs = _get_allocations_by_provider_id(context, rp.id)
    # Build up a list of Allocation objects, setting the Allocation object
    # fields to the same-named database record field we got from
    # _get_allocations_by_provider_id(). We already have the
    # ResourceProvider object so we just pass that object to the Allocation
    # object constructor as-is
    objs = []
    for rec in db_allocs:
        consumer = consumer_obj.Consumer(
            context, id=rec['consumer_id'],
            uuid=rec['consumer_uuid'],
            generation=rec['consumer_generation'],
            project=project_obj.Project(
                context, id=rec['project_id'],
                external_id=rec['project_external_id']),
            user=user_obj.User(
                context, id=rec['user_id'],
                external_id=rec['user_external_id']))
        objs.append(
            Allocation(
                id=rec['id'], resource_provider=rp,
                resource_class=context.rc_cache.string_from_id(
                    rec['resource_class_id']),
                consumer=consumer,
                used=rec['used'],
                created_at=rec['created_at'],
                updated_at=rec['updated_at']))
    return objs


def get_all_by_consumer_id(context, consumer_id):
    db_allocs = _get_allocations_by_consumer_uuid(context, consumer_id)

    if not db_allocs:
        return []

    # Build up the Consumer object (it's the same for all allocations
    # since we looked up by consumer ID)
    db_first = db_allocs[0]
    consumer = consumer_obj.Consumer(
        context, id=db_first['consumer_id'],
        uuid=db_first['consumer_uuid'],
        generation=db_first['consumer_generation'],
        consumer_type_id=db_first['consumer_type_id'],
        project=project_obj.Project(
            context, id=db_first['project_id'],
            external_id=db_first['project_external_id']),
        user=user_obj.User(
            context, id=db_first['user_id'],
            external_id=db_first['user_external_id']))

    # Build up a list of Allocation objects, setting the Allocation object
    # fields to the same-named database record field we got from
    # _get_allocations_by_consumer_id().
    #
    # NOTE(jaypipes):  Unlike with get_all_by_resource_provider(), we do
    # NOT already have the ResourceProvider object so we construct a new
    # ResourceProvider object below by looking at the resource provider
    # fields returned by _get_allocations_by_consumer_id().
    alloc_list = [
        Allocation(
            id=rec['id'],
            resource_provider=rp_obj.ResourceProvider(
                context,
                id=rec['resource_provider_id'],
                uuid=rec['resource_provider_uuid'],
                name=rec['resource_provider_name'],
                generation=rec['resource_provider_generation']),
            resource_class=context.rc_cache.string_from_id(
                rec['resource_class_id']),
            consumer=consumer,
            used=rec['used'],
            created_at=rec['created_at'],
            updated_at=rec['updated_at'])
        for rec in db_allocs
    ]
    return alloc_list


def replace_all(context, alloc_list):
    """Replace the supplied allocations.

    :note: This method always deletes all allocations for all consumers
           referenced in the list of Allocation objects and then replaces
           the consumer's allocations with the Allocation objects. In doing
           so, it will end up setting the Allocation.id attribute of each
           Allocation object.
    """
    # Retry _set_allocations server side if there is a
    # ResourceProviderConcurrentUpdateDetected. We don't care about
    # sleeping, we simply want to reset the resource provider objects
    # and try again. For sake of simplicity (and because we don't have
    # easy access to the information) we reload all the resource
    # providers that may be present.
    retries = context.config.placement.allocation_conflict_retry_count
    while retries:
        retries -= 1
        try:
            _set_allocations(context, alloc_list)
            break
        except exception.ResourceProviderConcurrentUpdateDetected:
            LOG.debug('Retrying allocations write on resource provider '
                      'generation conflict')
            # We only want to reload each unique resource provider once.
            alloc_rp_uuids = set(
                alloc.resource_provider.uuid for alloc in alloc_list)
            seen_rps = {}
            for rp_uuid in alloc_rp_uuids:
                # NOTE(melwitt): We use a separate database transaction to read
                # the resource provider because we might be wrapped in an outer
                # database transaction when we reach here. We want to get an
                # up-to-date generation value in case a racing request has
                # changed it after we began an outer transaction and this is
                # the first time we are reading the resource provider records
                # during our transaction.
                db_context_manager = db_api.placement_context_manager
                with db_context_manager.reader.independent.using(context):
                    seen_rps[rp_uuid] = rp_obj.ResourceProvider.get_by_uuid(
                        context, rp_uuid)
            for alloc in alloc_list:
                rp_uuid = alloc.resource_provider.uuid
                alloc.resource_provider = seen_rps[rp_uuid]
    else:
        # We ran out of retries so we need to raise again.
        # The log will automatically have request id info associated with
        # it that will allow tracing back to specific allocations.
        # Attempting to extract specific consumer or resource provider
        # information from the allocations is not coherent as this
        # could be multiple consumers and providers.
        LOG.warning('Exceeded retry limit of %d on allocations write',
                    context.config.placement.allocation_conflict_retry_count)
        raise exception.ResourceProviderConcurrentUpdateDetected()


def delete_all(context, alloc_list):
    consumer_uuids = set(alloc.consumer.uuid for alloc in alloc_list)
    alloc_ids = [alloc.id for alloc in alloc_list]
    _delete_allocations_by_ids(context, alloc_ids)
    consumer_obj.delete_consumers_if_no_allocations(
        context, consumer_uuids)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import os_resource_classes as orc
from oslo_serialization import jsonutils
from oslo_utils.fixture import uuidsentinel as uuids

from placement import direct
from placement import exception
from placement.objects import project as project_obj
from placement.tests.functional import base


class TestAllocationProjectCreateRace(base.TestCase):
    """Test that two allocation update request racing to create the project in
    the database. This test is added to reproduce the bug
    https://storyboard.openstack.org/#!/story/2009159 where the transaction
    that lost the project creation race fails as it tries to read the created
    project in the same transaction which is inactive due to the previous
    'Duplicate entry' error.
    """

    def setUp(self):
        super(TestAllocationProjectCreateRace, self).setUp()

        # Create resource provider and inventory for tests
        conf = self.conf_fixture.conf
        rp_data = jsonutils.dump_as_bytes({
            'name': 'a provider',
            'uuid': uuids.rp,
        })
        inv_data = jsonutils.dump_as_bytes({
            'inventories': {
                orc.VCPU: {
                    'total': 5,
                }
            },
            'resource_provider_generation': 0,
        })
        self.headers = {
            'x-auth-token': 'admin',
            'content-type': 'application/json',
            'OpenStack-API-Version': 'placement 1.38',
            'X_ROLES': 'admin,service'
        }
        with direct.PlacementDirect(conf) as client:
            # Create a resource provider
            url = '/resource_providers'
            resp = client.post(url, data=rp_data, headers=self.headers)
            self.assertEqual(200, resp.status_code)

            # Add inventory to the resource provider
            url = '/resource_providers/%s/inventories' % uuids.rp
            resp = client.put(url, data=inv_data, headers=self.headers)
            self.assertEqual(200, resp.status_code)

        # simulate that when the below allocation update call tries to fetch
        # the project it gets ProjectNotFound but at the same time a
        # "parallel" transaction creates the project, so the project creation
        # will fail
        real_get_project = project_obj.Project.get_by_external_id

        def fake_get_project(cls, ctx, external_id):
            if not hasattr(fake_get_project, 'called'):
                proj = project_obj.Project(ctx, external_id=external_id)
                proj.create()
                fake_get_project.called = True
                raise exception.ProjectNotFound(external_id)
            else:
                return real_get_project(ctx, external_id)

        self.useFixture(
            fixtures.MonkeyPatch(
                'placement.objects.project.Project.get_by_external_id',
                fake_get_project)
        )

    def test_set_allocations_for_consumer(self):
        alloc_data = jsonutils.dump_as_bytes({
            'allocations': {
                uuids.rp: {
                    'resources': {
                        orc.VCPU: 1,
                    },
                }
            },
            'project_id': uuids.project,
            'user_id': uuids.user,
            'consumer_generation': None,
            'consumer_type': 'INSTANCE',
        })
        conf = self.conf_fixture.conf
        with direct.PlacementDirect(conf) as client:
            # Create allocations
            url = '/allocations/%s' % uuids.consumer
            resp = client.put(url, data=alloc_data, headers=self.headers)

            # https://storyboard.openstack.org/#!/story/2009159 The expected
            # behavior would be that the allocation update succeeds as the
            # transaction can fetch the Project created by a racing transaction
            self.assertEqual(204, resp.status_code)

    def test_set_allocations(self):
        alloc_data = jsonutils.dump_as_bytes({
            uuids.consumer: {
                'project_id': uuids.project,
                'user_id': uuids.user,
                'consumer_generation': None,
                'consumer_type': 'INSTANCE',
                'allocations': {
                    uuids.rp: {
                        'resources': {
                            orc.VCPU: 1,
                        },
                    }
                }
            }
        })
        conf = self.conf_fixture.conf
        with direct.PlacementDirect(conf) as client:
            # Create allocations
            url = '/allocations'
            resp = client.post(url, data=alloc_data, headers=self.headers)
            self.assertEqual(204, resp.status_code)

    def test_reshape(self):
        alloc_data = jsonutils.dump_as_bytes({
            'allocations': {
                uuids.consumer: {
                    'allocations': {
                        uuids.rp: {
                            'resources': {
                                orc.VCPU: 1,
                            },
                        }
                    },
                    'project_id': uuids.project,
                    'user_id': uuids.user,
                    'consumer_generation': None,
                    'consumer_type': 'INSTANCE',
                }
            },
            'inventories': {
                uuids.rp: {
                    'inventories': {
                        orc.VCPU: {
                            'total': 5,
                        }
                    },
                    'resource_provider_generation': 1,
                }
            }
        })
        conf = self.conf_fixture.conf
        with direct.PlacementDirect(conf) as client:
            # Create allocations
            url = '/reshaper'
            resp = client.post(url, data=alloc_data, headers=self.headers)
            self.assertEqual(204, resp.status_code)



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_db import options as oslo_db_options

_ENRICHED = False


def enrich_help_text(alt_db_opts):

    def get_db_opts():
        for group_name, db_opts in oslo_db_options.list_opts():
            if group_name == 'database':
                return db_opts
        return []

    for db_opt in get_db_opts():
        for alt_db_opt in alt_db_opts:
            if alt_db_opt.name == db_opt.name:
                # NOTE(markus_z): We can append alternative DB specific help
                # texts here if needed.
                alt_db_opt.help = db_opt.help + alt_db_opt.help


# NOTE(markus_z): We cannot simply do:
# conf.register_opts(oslo_db_options.database_opts, 'placement_database')
# If we reuse a db config option for two different groups ("placement_database"
# and "database") and deprecate or rename a config option in one of these
# groups, "oslo.config" cannot correctly determine which one to update.
# That's why we copied & pasted these config options for the
# "placement_database" group here. See nova commit ba407e3 ("Add support
# for multiple database engines") for more details.
# TODO(cdent): Consider our future options of using 'database' instead of
# 'placement_database' for the group. This is already loose in the wild,
# explicit, and safe if there will ever be more than one database, so may
# be good to leave it.
placement_db_group = cfg.OptGroup('placement_database',
                                  title='Placement API database options',
                                  help="""
The *Placement API Database* is a the database used with the placement
service. If the connection option is not set, the placement service will
not start.
""")

placement_db_opts = [
    cfg.StrOpt(
        'connection',
        help='',
        required=True,
        secret=True),
    cfg.StrOpt(
        'connection_parameters',
        default='',
        help=''),
    cfg.BoolOpt(
        'sqlite_synchronous',
        default=True,
        help=''),
    cfg.StrOpt(
        'slave_connection',
        secret=True,
        help=''),
    cfg.StrOpt(
        'mysql_sql_mode',
        default='TRADITIONAL',
        help=''),
    cfg.IntOpt(
        'connection_recycle_time',
        default=3600,
        help=''),
    cfg.IntOpt(
        'max_pool_size',
        help=''),
    cfg.IntOpt(
        'max_retries',
        default=10,
        help=''),
    cfg.IntOpt(
        'retry_interval',
        default=10,
        help=''),
    cfg.IntOpt(
        'max_overflow',
        help=''),
    cfg.IntOpt(
        'connection_debug',
        default=0,
        help=''),
    cfg.BoolOpt(
        'connection_trace',
        default=False,
        help=''),
    cfg.IntOpt(
        'pool_timeout',
        help=''),
    cfg.BoolOpt(
        'sync_on_startup',
        default=False,
        help='If True, database schema migrations will be attempted when the'
             ' web service starts.'),
]


def register_opts(conf):
    conf.register_opts(placement_db_opts, group=placement_db_group)


def list_opts():
    # NOTE(markus_z): 2016-04-04: If we list the oslo_db_options here, they
    # get emitted twice(!) in the "sample.conf" file. First under the
    # namespace "nova.conf" and second under the namespace "oslo.db". This
    # is due to the setting in file "etc/nova/nova-config-generator.conf".
    # As I think it is useful to have the "oslo.db" namespace information
    # in the "sample.conf" file, I omit the listing of the "oslo_db_options"
    # here.
    global _ENRICHED
    if not _ENRICHED:
        enrich_help_text(placement_db_opts)
        _ENRICHED = True
    return {
        placement_db_group: placement_db_opts,
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Extend functionality from webob.dec.wsgify for Placement API."""

import webob

from oslo_log import log as logging
from webob.dec import wsgify

from placement import util

LOG = logging.getLogger(__name__)


class PlacementWsgify(wsgify):

    def call_func(self, req, *args, **kwargs):
        """Add json_error_formatter to any webob HTTPExceptions."""
        try:
            super(PlacementWsgify, self).call_func(req, *args, **kwargs)
        except webob.exc.HTTPException as exc:
            LOG.debug("Placement API returning an error response: %s", exc)
            exc.json_formatter = util.json_error_formatter
            # The exception itself is not passed to json_error_formatter
            # but environ is, so set the environ.
            if exc.comment:
                req.environ[util.ENV_ERROR_CODE] = exc.comment
                exc.comment = None
            raise



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools
import prettytable
import sys

from oslo_config import cfg
from oslo_log import log as logging
import pbr.version

from placement import conf
from placement import context
from placement.db.sqlalchemy import migration
from placement import db_api
from placement.objects import consumer as consumer_obj
from placement.objects import resource_provider as rp_obj

version_info = pbr.version.VersionInfo('openstack-placement')
LOG = logging.getLogger(__name__)

online_migrations = (
    # These functions are called with a DB context and a count, which is the
    # maximum batch size requested by the user. They must be idempotent.
    # At most $count records should be migrated. The function must return a
    # tuple of (found, done). The found value indicates how many
    # unmigrated/candidate records existed in the database prior to the
    # migration (either total, or up to the $count limit provided), and a
    # nonzero found value may tell the user that there is still work to do.
    # The done value indicates whether or not any records were actually
    # migrated by the function. Thus if both (found, done) are nonzero, work
    # was done and some work remains. If found is nonzero and done is zero,
    # some records are not migratable, but all migrations that can complete
    # have finished.

    # Added in Stein
    rp_obj.set_root_provider_ids,
    # Added in Stein (copied from migration added to Nova in Rocky)
    consumer_obj.create_incomplete_consumers,
)


class DbCommands(object):
    def __init__(self, config):
        self.config = config

    def db_sync(self):
        # Let exceptions raise for now, they will go to stderr.
        migration.upgrade('head')
        return 0

    def db_version(self):
        print(migration.version())
        return 0

    def db_stamp(self):
        migration.stamp(self.config.command.version)
        return 0

    def db_online_data_migrations(self):
        """Processes online data migration.

        :returns: 0 if no (further) updates are possible, 1 if the
                  ``--max-count`` option was used and some updates were
                  completed successfully (even if others generated errors),
                  2 if some updates generated errors and no other migrations
                  were able to take effect in the last batch attempted, or
                  127 if invalid input is provided.
        """
        max_count = self.config.command.max_count
        if max_count is not None:
            try:
                max_count = int(max_count)
            except ValueError:
                max_count = -1
            if max_count < 1:
                print('Must supply a positive value for max_count')
                return 127
            limited = True
        else:
            max_count = 50
            limited = False
            print('Running batches of %i until complete' % max_count)

        ran = None
        migration_info = collections.OrderedDict()
        exceptions = False
        while ran is None or ran != 0:
            migrations, exceptions = self._run_online_migration(max_count)
            ran = 0
            # For each batch of migration method results, build the cumulative
            # set of results.
            for name in migrations:
                migration_info.setdefault(name, (0, 0))
                migration_info[name] = (
                    migration_info[name][0] + migrations[name][0],
                    migration_info[name][1] + migrations[name][1],
                )
                ran += migrations[name][1]
            if limited:
                break

        t = prettytable.PrettyTable(
            ['Migration', 'Total Found', 'Completed'])
        for name, info in migration_info.items():
            t.add_row([name, info[0], info[1]])
        print(t)

        # NOTE(tetsuro): In "limited" case, if some update has been "ran",
        # exceptions are not considered fatal because work may still remain
        # to be done, and that work may resolve dependencies for the failing
        # migrations.
        if exceptions and not (limited and ran):
            print("Some migrations failed unexpectedly. Check log for "
                  "details.")
            return 2

        # TODO(mriedem): Potentially add another return code for
        # "there are more migrations, but not completable right now"
        return ran and 1 or 0

    def _run_online_migration(self, max_count):
        ctxt = context.RequestContext(config=self.config)
        ran = 0
        exceptions = False
        migrations = collections.OrderedDict()
        for migration_meth in online_migrations:
            count = max_count - ran
            try:
                found, done = migration_meth(ctxt, count)
            except Exception:
                msg = ("Error attempting to run %(method)s" % dict(
                    method=migration_meth))
                print(msg)
                LOG.exception(msg)
                exceptions = True
                found = done = 0

            name = migration_meth.__name__
            if found:
                print('%(total)i rows matched query %(meth)s, %(done)i '
                      'migrated' % {'total': found,
                                    'meth': name,
                                    'done': done})
            # This is the per-migration method result for this batch, and
            # _run_online_migration will either continue on to the next
            # migration, or stop if up to this point we've processed max_count
            # of records across all migration methods.
            migrations[name] = found, done
            ran += done
            if ran >= max_count:
                break
        return migrations, exceptions


def add_db_command_parsers(subparsers, config):
    command_object = DbCommands(config)

    # If we set False here, we avoid having an exit during the parse
    # args part of CONF processing and we can thus print out meaningful
    # help text.
    subparsers.required = False
    parser = subparsers.add_parser('db')
    # Avoid https://bugs.python.org/issue9351 with cpython < 2.7.9
    parser.set_defaults(func=parser.print_help)
    db_parser = parser.add_subparsers(description='database commands')

    help = 'Sync the database to the current version.'
    sync_parser = db_parser.add_parser('sync', help=help, description=help)
    sync_parser.set_defaults(func=command_object.db_sync)

    help = 'Report the current database version.'
    version_parser = db_parser.add_parser(
        'version', help=help, description=help)
    version_parser.set_defaults(func=command_object.db_version)

    help = 'Stamp the revision table with the given version.'
    stamp_parser = db_parser.add_parser('stamp', help=help, description=help)
    stamp_parser.add_argument('version', help='the version to stamp')
    stamp_parser.set_defaults(func=command_object.db_stamp)

    help = 'Run the online data migrations.'
    online_dm_parser = db_parser.add_parser(
        'online_data_migrations', help=help, description=help)
    online_dm_parser.add_argument(
        '--max-count', metavar='<number>',
        help='Maximum number of objects to consider')
    online_dm_parser.set_defaults(
        func=command_object.db_online_data_migrations)


def setup_commands(config):
    # This is a separate method because it facilitates unit testing.
    # Use an additional SubCommandOpt and parser for each new sub command.
    add_db_cmd_parsers = functools.partial(
        add_db_command_parsers, config=config)
    command_opt = cfg.SubCommandOpt(
        'db', dest='command', title='Command', help='Available DB commands',
        handler=add_db_cmd_parsers)
    return [command_opt]


def main():
    config = cfg.ConfigOpts()
    conf.register_opts(config)
    command_opts = setup_commands(config)
    config.register_cli_opts(command_opts)
    config(sys.argv[1:], project='placement',
           version=version_info.version_string(),
           default_config_files=None)
    db_api.configure(config)

    try:
        func = config.command.func
        return_code = func()
        # If return_code ends up None we assume 0.
        sys.exit(return_code or 0)
    except cfg.NoSuchOptError:
        config.print_help()
        sys.exit(1)
